# __revision__  = $LastChangedRevision$
# __Copyright__ = Copyright (c) 2007 - 2009, CollabNet, Inc. All rights reserved.

from ConfigParser import ConfigParser
import sys
import time

CONTEXT_NAMES = ['group', 'package', 'page', 'planningfolder', 'project', 'release',
                 'tracker', 'user', 'wikipage', 'category']

Indices  = ['docfolders', 'groups', 'pages', 'planningfolders', 'projects', 'users']

def get_tzname():
    now = time.localtime()
    isdst = now[8]
    if isdst :
        return time.tzname[1]
    else:
        return time.tzname[0]

#TODO: replace individual objects with a generic TFObject class
class Project(object):
    """Create a project object from a xml data"""
    def __init__(self, data):
        self.id = data.find('id').text
        self.path = data.find('path').text
        self.name = self.path.split('.')[1]
        self.description = data.find('description').text
        self.title = data.find('title').text
        self.xml = data

class User(object):

    def __init__(self, data):
        self.id = data.find('id').text
        self.username = data.find('username').text
        self.email = data.find('email').text
        self.fullName = data.find('fullName').text
        self.superUser = data.find('superUser').text
        self.restrictedUser = data.find('restrictedUser').text
        self.xml = data

class Group(object):
    def __init__(self, data):
        self.id = data.find('id').text
        self.fullName = data.find('fullName').text
        self.description = data.find('description').text
        self.xml = data

class Role(object):
    def __init__(self, data):
        self.id = data.find('id').text
        self.title = data.find('title').text
        self.description = data.find('description').text
        self.xml = data

class Page(object):
    def __init__(self, data):
        self.id = data.find('id').text
        self.description = data.find('description').text
        self.title = data.find('title').text
        self.xml = data

class Tracker(object):
    def __init__(self, data):
        self.id = data.find('id').text
        self.description = data.find('description').text
        self.title = data.find('title').text
        self.xml = data

class Field(object):
    def __init__(self, data):
        self.id = data.find('id').text
        self.fieldType = data.find('fieldType').text
        self.name = data.find('name').text
        self.xml = data

class PlanningFolder(object):
    def __init__(self, data):
        self.id = data.find('id').text
        self.description = data.find('description').text
        self.title = data.find('title').text
        self.xml = data

class WikiPage(object):
    def __init__(self, data):
        self.id = data.find('id').text
        self.title = data.find('title').text
        self.xml = data

class Release(object):
    def __init__(self, data):
        self.id = data.find('id').text
        self.title = data.find('title').text
        self.xml = data

class Package(object):
    def __init__(self, data):
        self.id = data.find('id').text
        self.title = data.find('title').text
        self.xml = data

class Category(object):
    def __init__(self, data):
        self.id = data.find('id').text
        self.title = data.find('title').text
        self.xml = data

class Respository(object):
    def __init__(self, data):
        self.id = data.find('id').text
        self.title = data.find('title').text
        self.xml = data

class Session(object):
    def __init__(self):
        self.Ctx = {}
        self.app_version = None
        self.api_version = None
        self.backupdir = None
        self.base_url = 'http://localhost:8080'
        self.cache = ConfigParser()
        self.error = 0
        self.error_code = None
        self.errmsg = None
        self.exportdata = False
        self.help = False
        self.interactive = False
        self.isSuperUser = False
        self.isRestrictedUser = False
        self.locale = 'en'
        self.logfile = sys.stderr
        self.loglevel = 'INFO'
#        self.mc = memcache.Client([MEMCACHEADDR], debug=0)
        self.format = 'table'
        self.outfile = None
        self.passwd = None
        self.session_id = None
        self.soaptrace = None
        # valid form: "PST", "America/Los_Angeles", "GMT-8:00"
        self.timeZone = get_tzname()
        self.min_version = '5.2.0.0'
        self.isCTF53 = False
        self.tar = None
        self.user = None
        self.verbose = False
        self.kw_soapopt = {}
        self.Apps_url = {}
        self.Version = None
        self.ApiVersion = None
        self.trackerFields = {}       # k,v = id, etree.Element node
        self.wikiPages = {}           # k,v = title, WikiPageSoapDO
        self.artifacts = {}           # k,v = id, etree.Element node
        self.indices = {}
        self.index_last_updated = {}
        for i in Indices:
            self.indices[i] = {}
 
        self.SoapApps = ['AdminService', 'CategorizationApp', 'CollabNet', 'DiscussionApp',
        		 'DocumentApp', 'FileStorageApp', 'FrsApp', 'IntegrationDataApp',
        		 'NewsApp', 'PageApp', 'RbacApp', 'ScmApp', 'SimpleFileStorageApp',
        		 'TaskApp', 'TrackerApp', 'Version', 'WikiApp', 'PlanningApp']

        # these attributes can be updated with 'Set' command:
        self.RW_attributes = ['backupdir',
                              'exportdata',
                              'format',
                              'interactive',
                              'isRestrictedUser',
                              'isSuperUser',
                              'locale',
                              'logfile',
                              'loglevel',
                              'outfile',
                              'session_id',
                              'soaptrace',
                              'timeZone',
                              'verbose',
                              'Version',
                              'ApiVersion',
                             ]
    
        self.set_app_url()
       
    def set_app_url(self):
        if self.base_url:
            for app in self.SoapApps:
                #self.Apps_url[app] = self.base_url + \
                #                            '/ce-soap60/services/' + app
                self.Apps_url[app] = "%s/ce-soap60/services/%s" % (self.base_url, app)
        else:
            for app in self.SoapApps:
                self.Apps_url[service] = 'http://localhost:8080/ce-soap60/services/' + app

    def get_app_url(self, app):
        if app in self.SoapApps:
            return self.Apps_url[app]
        else:
            return None

#    def cache_login_info(self):
#        if self.session_id is None:
#            return
#
#    def setup_cache(self):
#        home = os.environ['HOME']
#        rcfile = os.path.join(home, '.teamforgerc')
#        if os.path.exists(rcfile):
#            try:
#                f = open(rcfile, 'r')
#                self.cache.readfp(f)
#                f.close()
#            except Exception, err:
#                raise IOError, "Could not read %s: %s" % (rcfile, err)
