
# __revision__  = $LastChangedRevision$
# __Copyright__ = Copyright (c) 2007 - 2009, CollabNet, Inc. All rights reserved.

import cmd
import os
import sys
#from copy import deepcopy
from lxml import etree

BASEDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
sys.path.append(os.path.join(BASEDIR, "TeamForge"))

import LineParser
from ImportModule import Commands
from ImportModule import SoapApps
from ImportModule import ApiMethods
from ApiHelpDocs  import HELP_SoapData

import curses
curses.setupterm()
MAX_WIDTH = curses.tigetnum('cols')

# default connection info:
DEFAULT_URL = 'http://localhost:8080'
DEFAULT_USER = 'admin'
DEFAULT_PASSWD = 'admin'

#(Commands, SoapApps, ApiMethods) = ImportModule.mapCommands('Commands')
# for debugging only
#print "  %s => %s" % ('Command Name'.center(25), 'Class reference')
#for k in Commands.keys():
#    print "  %s => %s" % (k.ljust(25), Commands[k])
#print "  %s => %s" % ('Api Name'.center(25), 'Class reference')
#for k in ApiMethods.keys():
#    print "  %s => %s" % (k.ljust(25), ApiMethods[k])
#print "  %s => %s" % ('Api SoapApps'.center(25), 'Api Methods')
#for k in SoapApps.keys():
#    print "  %s => %s" % (k.ljust(25), SoapApps[k])

class CmdProcessor(cmd.Cmd):
    """CTF Command Processor"""

    session = None
    prompt = ''

    def __init__(self, session, completekey='tab', stdin=None, stdout=None):
        cmd.Cmd.__init__(self, completekey='tab', stdin=None, stdout=None)
        self.session = session
        if self.session.interactive:
            CmdProcessor.prompt = 'APItool> '
        else:
            CmdProcessor.prompt = ''

        self.connect()
        if self.session.session_id is not None:
            CMD = 'getVersion'
            self.session.Version = ApiMethods[CMD](self.session, CMD, []).export_data()
            CMD = 'getApiVersion'
            self.session.ApiVersion = ApiMethods[CMD](self.session, CMD, []).export_data()

    def connect(self):
        if self.session.base_url:
            Commands['server'](self.session, [self.session.base_url]).run()
        else:
            Commands['server'](self.session, [DEFAULT_URL]).run()
            if self.session.error == 0:
                self.session.base_url = DEFAULT_URL

        if self.session.user and self.session.passwd:
            ApiMethods['login'](self.session, 'login',
                        [self.session.user, self.session.passwd ]).run()
        else:
            ApiMethods['login'](self.session, 'login',
                        [DEFAULT_USER, DEFAULT_PASSWD]).run()
            if self.session.error == 0:
                self.session.user = DEFAULT_USER
                self.session.passwd = DEFAULT_PASSWD

    def default(self, line):
        # skip comment lines
        if line.strip().startswith('#'):
            return
        tokens = LineParser.parse(line)
        this_cmd = tokens[0]
        arg_list = tokens[1:]
        if this_cmd == 'quit':
            Commands[this_cmd]()
        elif this_cmd in Commands:
            Commands[this_cmd](self.session, arg_list).run()
        elif this_cmd in ApiMethods:
            ApiMethods[this_cmd](self.session, this_cmd, arg_list).run()
        else:
            print >> sys.stderr, "ERROR: unknown command: %s" % this_cmd

#    def precmd(self, line):
#        pass

#    def postloop(self):
#        pass

    def do_shell(self, line):
        "usage: Run a shell command"
        output = os.popen(line).read()
        print >> sys.stderr, output
        self.last_output = output

    def do_help(self, line):
        self.session.help = True
        if line:
            this_cmd = LineParser.parse(line)[0]
            if this_cmd and this_cmd in Commands:
                Commands[this_cmd](self.session, this_cmd).usage()
            elif this_cmd and this_cmd in ApiMethods:
                ApiMethods[this_cmd](self.session, this_cmd).usage()
            elif this_cmd and this_cmd in SoapApps:
                cmd_list = SoapApps[this_cmd]
                self.print_cmdlist(cmd_list, 
                     "  Helps are available for following Soap services from %s:" % this_cmd)
            elif this_cmd and this_cmd in HELP_SoapData:
                import pydoc
                pydoc.ttypager(HELP_SoapData[this_cmd])
            else:
                print >> sys.stderr, "ERROR: unknown, no help on '%s'" % this_cmd
                
        else :
            cmd_list = Commands.keys()
            self.print_cmdlist(cmd_list, 
                 "Helps are available for following commands:")

            cmd_list = SoapApps.keys()
            self.print_cmdlist(cmd_list, 
                 "Helps are available for following Soap Applications:")

            datatype_list = HELP_SoapData.keys()
            self.print_cmdlist(datatype_list,
                 "Helps are available for following ComplexType SOAP data objects:")

        self.session.help = False

    def print_cmdlist(self, clist, msg):

        clist.sort()
        num_cmds = len(clist)
        max_len = 0
        for c in clist:
            if len(c) > max_len:
                max_len = len(c)
        col_width = max_len + 2
        ncol = MAX_WIDTH / col_width
        nrow = num_cmds / ncol 
        if num_cmds % ncol > 0 :
            nrow += 1

        print >> sys.stderr, "  %s" % msg
        for n in range(nrow):
            line = '    '
            for col in range(ncol):
                N = col * nrow + n
                if N < num_cmds:
                    line += clist[N].ljust(col_width)
            print >> sys.stderr, line
        print >> sys.stderr, ""

    # when reading input from a file, this causes clean exit at EOF
    def do_EOF(self, line):
        if self.session.session_id:
            ApiMethods['logoff'](self.session, 'logoff', [self.session.session_id]).run()
        return True

    # override default behavior
    def emptyline(self):
        return

