#!/usr/bin/env python

import sys

# __revision__  = $LastChangedRevision$
# __Copyright__ = Copyright (c) 2007 - 2009, CollabNet, Inc. All rights reserved.


def parse(line, delimiter=None):
    """
    This function parses a string into tokens. If a token contains
    white space, it should be quoted with either single or double quote
    """
    tokens = []
    # set up some flags and indicies
    t = t_s = t_e = sq = dq = 0

    if delimiter:
        d = delimiter
    else:
        d = [' ', '\t', '|']
    # delete enverything after first #, and strip spaces
    c = line.find('#')
    if c > 0:
        line = line[0:c].strip()
    else:
        line = line.strip()

    for pos in range(len(line)):
        c = line[pos]
        if sq == 0 and dq == 0 and t == 0 :
            if c in d :
                pass
            elif c == '\'' :
                t = 1
                sq = 1
                t_s = pos
            elif c == '\"' :
                t = 1
                dq = 1
                t_s = pos
            else:
                t = 1
                t_s = pos
        elif sq == 0 and dq == 0 and t == 1 :
            if c in d:
                t = 0
                t_e = pos
                s = line[t_s:t_e].strip('"\'')
                #tokens.append(line[t_s:t_e].strip('"\''))  # remove leading and trailing quotes
                tokens.append(s)  # remove leading and trailing quotes
            elif c == '"':
                pass
            else:
                pass
        elif sq == 0 and dq == 1 and t == 0 :
            #illegal combo
            print "ERROR: LineParser(), illegal states"
        elif sq == 0 and dq == 1 and t == 1 :
            if c == '"' :
                dq = 0
            else:
                pass
        elif sq == 1 and dq == 0 and t == 0 :
            print "ERROR: LineParser(), illegal states"
            sys.exit(1)
        elif sq == 1 and dq == 0 and t == 1 :
            if c == '\'' :
                sq = 0
            elif c == '"':
                pass
            else:
                pass
    if t == 1:
        tokens.append(line[t_s:].strip('"\''))
    return tokens

if __name__ == "__main__" :

    # test doc:
    test_doc = """
    normal tokens "John Doe"
    line with '' token or " " token
    My name is  O'keefe

    This	is a " token with 'single quotes'"
    admin admin 	'the admin' a=b
      # comment line """

    lines = test_doc.split('\n')
    for l in lines:
        print l
        print parse(l)

# Note:
# Truth table for LineParser
#     sq = single quote, dq = double quote, 
#     t = True|False is if we already have a token started
#     d = delimiter, currently is space,
#
# case	prev. state           new char
#-------------------------     --------
#	sq	dq	t	C	transition	desc.
#------------------------------------------------------------------------
# 1	-	-	-	d			consecutive delimiters
# 2	-	-	-	sq	t:T,sq:1	\
# 3	-	-	-	dq	t:T,dq:1	 new token starts
# 4	-	-	-	c	t:T,		/
#	
# 5	-	-	+	d	t:F, t_e=post	end of token, got a complete token
# 6	-	-	+	sq			\	
# 6	-	-	+	dq			 continueing token
# 6	-	-	+	c			/
#  	
# 7	-	+	-	d	illegal state	if one of sq|dq is set, t_start must be True!
# 7	-	+	-	sq	illegal state
# 7	-	+	-	dq	illegal state
# 7	-	+	-	c	illegal state
#	
# 6	-	+	+	d			\
# 6	-	+	+	sq			 \ continueing token
# 8	-	+	+	dq	dq:0,q:0		 /
# 6	-	+	+	c			/
#	
# 7	+	-	-	d	illegal state	if sq|dq is set, t must be True!
# 7	+	-	-	sq
# 7	+	-	-	dq
# 7	+	-	-	c
#	
# 6	+	-	+	d			\
# 9	+	-	+	sq	sq:0		 \ continueing token
# 6	+	-	+	dq			 /
# 6	+	-	+	c			/
#
