
# __revision__  = $LastChangedRevision$
# __Copyright__ = Copyright (c) 2007 - 2009, CollabNet, Inc. All rights reserved.

from lxml import etree
from ImportModule import ApiMethods
from SimpleLog import SimpleLog
import DataUtils
import os
import random
import sys

commands = ['exportUsers', 'importUsers', 'exportGroups', 'importGroups'] 

def randString(size, base=None):
    bases = {'numbers' : '01234567890123456789012345678901234567890123456789',
             'letters' : 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ',
             'lower'   : 'abcdefghijklmnopqrstuvwxyz',
             'lowernum'   : 'abcdefghijklmnopqrstuvwxyz0123456789',
             'upper'   : 'ABCDEFGHIJKLMNOPQRSTUVWXYZ',
             'uppernum'   : 'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789',
             None      : 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789',
            }
    if base in bases :
        return ''.join(random.sample(bases[base], size))
    else:
        return ''.join(random.sample(base, size))

class exportUsers(object):
    def __init__(self, session, args=None):
        self.session = session
        self.args = args
        self.input = None
        self.log = SimpleLog(self.session.loglevel)
        self.users = None

    def usage(self, errmsg=None):
        if errmsg:
            self.log.error("%s" % errmsg)
        self.log.info("export all users data and save results to a file")
        self.log.info("Usage: exportUsers filename")

    def validate(self):
        if self.session.session_id is None:
            self.log.error("Active session not found, abort")
            return

        if self.args and len(self.args) != 1:
            self.usage("syntax error, missing filename")
            return False

        if os.path.exists(self.args[0]):
            self.log.error("file '%s' exist, either delete it or give another name" % self.args[0])
            return False
        else:
            self.input = self.args[0]

        return True

    def run(self):
        if not self.validate():
            return
        f = DataUtils.create_archive(self.input)   # this returns regular file object
        if f is None:
            return
        count = 0
        self.log.info("Exporting all users.")
        self.users = etree.Element('Users')
        userList = []
        cmd = 'getUserList'
        userList = ApiMethods[cmd](self.session, cmd, []).export_xml()
        if self.session.error > 0:
            self.log.error("failed to get user list")
            sys.exit(1)
        cmd = 'getUserData'
        for user in userList:
            status = user.find('status').text
            if status in ['Removed', 'Disabled']:
                continue
            userName = user.find('userName').text
            args = [userName]
            userData = ApiMethods[cmd](self.session, cmd, args).export_xml()
            if self.session.error > 0:
                self.log.error("failed to get user data for user '%s'" % userName)
            else:
                self.log.debug("  exporting user data for '%s'" % userName)
            if userData is not None:
                self.users.append(userData)
                count += 1
            else:
                self.log.error("exportUsers: unexpected error")
        print >> f, etree.tostring(self.users, 
                                   pretty_print=True, 
                                   encoding='utf-8', 
                                   xml_declaration=True)
        self.log.info("Total of %d users written to '%s'" % (count, self.input))
        self.log.info("exportUsers finished successfully")
        f.close()

class importUsers(object):
    def __init__(self, session, args=None):
        self.session = session
        self.args = args
        self.log = SimpleLog(self.session.loglevel)
        self.users = None
        self.existUsers = self.getExistUsers()

    def usage(self, errmsg=None):
        if errmsg:
            self.log.error("%s" % errmsg)
        self.log.info("import user accounts, skip if user account already exists.")
        self.log.info("Usage: importUsers filename")

    def validate(self):
        if self.session.session_id is None:
            self.log.error("Active session not found, abort")
            return

        if self.args and len(self.args) != 1:
            self.usage("syntax error, missing filename")
            return False

        if not os.path.exists(self.args[0]):
            self.log.error("file '%s' does not exist." % self.args[0])
            return False

        self.users = etree.parse(self.args[0]).getroot()
        return True

    def getExistUsers(self):
        """Get a list of all existing users"""
        cmd = 'getUserList'
        return ApiMethods[cmd](self.session, cmd, []).export_xml()

    def run(self):
        if not self.validate():
            return
        if self.users is None:
            self.log.warn("There are no users in input file")
            return
        else:
            count = 0
        self.log.info("Importing users")
        for user in self.users.iterfind('User'):
            userName = user.find('username').text
            # xpath return a list
            if self.existUsers.xpath("//User[username='%s']" % userName):
                self.log.debug("  skip user '%s', account exists" % userName)
                continue
            email = user.find('email').text
            fullName = user.find('fullName').text
            locale = user.find('locale').text
            # Organization, licenseType fields are validated to support importUsers to TF7.0 from previous releases
            if (user.find('organization') is not None):
                organization = user.find('organization').text
            else:
                organization = ''
            if (user.find('licenseType') is not None):
                licenseType = user.find('licenseType').text
            else:
                licenseType = 'ALM'
            restrictedUser = user.find('restrictedUser').text
            superUser = user.find('superUser').text
            timeZone = user.find('timeZone').text
            pwd = randString(10)
            cmd = 'createUser'
            args = [userName, email, fullName, organization, locale, timeZone, 
                    licenseType, superUser, restrictedUser, pwd ]
            userData = ApiMethods[cmd](self.session, cmd, args).export_xml()
            if self.session.error > 0 and \
               self.session.error_code == 'ObjectAlreadyExistsFault':
                self.log.debug("  skip user '%s', account exists" % userName)
            else:
                self.log.debug("  created account for user '%s'" % userName)
            if userData is not None:
                count += 1
        self.log.info("New accounts created: %s" % count)
        self.log.info("importUsers finished successfully")

class exportGroups(object):
    def __init__(self, session, args=None):
        self.session = session
        self.args = args
        self.input = None
        self.log = SimpleLog(self.session.loglevel)
        self.groups = None

    def usage(self, errmsg=None):
        if errmsg:
            self.log.error("%s" % errmsg)
        self.log.info("export groups and all their active members to a file")
        self.log.info("Usage: exportGroups filename")

    def validate(self):
        if self.session.session_id is None:
            self.log.error("Active session not found, abort")
            return

        if self.args and len(self.args) != 1:
            self.usage("syntax error, missing filename")
            return False

        if os.path.exists(self.args[0]):
            self.log.error("file '%s' exist, either delete it or give another name" % self.args[0])
            return False
        else:
            self.input = self.args[0]

        return True

    def run(self):
        if not self.validate():
            return
        # this returns regular file object
        f = DataUtils.create_archive(self.input)   
        if f is None:
            return
        count = 0
        self.log.info("Exporting all groups.")
        self.groups = etree.Element('Groups')
        groupList = []
        cmd = 'getUserGroupList'
        groupList = ApiMethods[cmd](self.session, cmd, []).export_xml()
        if self.session.error > 0:
            self.log.error("failed to get group list")
            sys.exit(1)
        cmd = 'getUserGroupData'
        for group in groupList:
            xid = group.find('id').text
            fullName = group.find('fullName').text
            args = [xid]
            groupData = ApiMethods[cmd](self.session, cmd, args).export_xml()
            if self.session.error > 0:
                self.log.error("failed to get group data for group '%s'" % fullName)
            else:
                self.log.debug("  exporting group data for '%s'" % fullName)
            if groupData is not None:
                groupData.append(self.exportGroupMembers(group))
                self.groups.append(groupData)
                count += 1
            else:
                self.log.error("exportGroups: unexpected error")
        print >> f, etree.tostring(self.groups, 
                                   pretty_print=True, 
                                   encoding='utf-8', 
                                   xml_declaration=True)
        self.log.info("Total of %d groups written to '%s'" % (count, self.input))
        self.log.info("exportGroups finished successfully")
        f.close()

    def exportGroupMembers(self, group):
        cmd = 'getUserGroupMembers'
        groupId = group.find('id').text
        fullName = group.find('fullName').text
        args = [groupId]
        members = ApiMethods[cmd](self.session, cmd, args).export_xml()
        if self.session.error > 0:
            self.log.error("failed to get group members for group '%s'" % fullName)
        else:
            self.log.debug("  exporting group members for group '%s'" % fullName)
        return members

class importGroups(object):
    def __init__(self, session, args=None):
        self.session = session
        self.args = args
        self.log = SimpleLog(self.session.loglevel)
        self.groups = None
        self.known_groups = self.getAllGroups()

    def usage(self, errmsg=None):
        if errmsg:
            self.log.error("%s" % errmsg)
        self.log.info("import groups , skip if group already exists.")
        self.log.info("Usage: importGroups filename")

    def validate(self):
        if self.session.session_id is None:
            self.log.error("Active session not found, abort")
            return

        if self.args and len(self.args) != 1:
            self.usage("syntax error, missing filename")
            return False

        if not os.path.exists(self.args[0]):
            self.log.error("file '%s' does not exist." % self.args[0])
            return False

        self.groups = etree.parse(self.args[0]).getroot()
        return True

    def getAllGroups(self):
        """Get a list of all existing groups"""
        cmd = 'getUserGroupList'
        return ApiMethods[cmd](self.session, cmd, []).export_xml()

    def getExistGroupMembers(self, groupId):
        cmd = "getUserGroupMembers"
        return ApiMethods[cmd](self.session, cmd, [groupId]).export_xml()

    def run(self):
        if not self.validate():
            return
        if self.groups is None:
            self.log.warn("There are no groups in input file")
            return
        else:
            count = 0
        existGroupId = None
        self.log.info("Importing groups")
        # Support importing of UserGroup from Soap60
        if self.groups.findtext('UserGroup'):
            groupObj         = 'UserGroup'
            userGroupListObj = 'UserGroupMembers'
            userGroupObj     = 'UserGroupMember'
        else:
            groupObj         = 'Group'
            userGroupListObj = 'ActiveGroupMembers'
            userGroupObj     = 'ActiveGroupMember'
        for group in self.groups.iterfind(groupObj):
            fullName = group.find('fullName').text
            # query from cached known_goups
            existGroupId = self.known_groups.xpath("//Group[fullName='%s']/id" % fullName)
            if existGroupId:
                self.log.debug("  skip group '%s', group already exists" % fullName)
                self.importGroupMembers(existGroupId[0].text, fullName, group.find(userGroupListObj), userGroupObj)
                continue
            description = group.find('description').text
            cmd = 'createUserGroup'
            args = [fullName, description ]
            newGroup = ApiMethods[cmd](self.session, cmd, args).export_xml()
            if self.session.error > 0 :
                self.log.error("  failed to create group '%s'" % fullName)
            else:
                self.log.debug("  created group '%s'" % fullName)
            if newGroup is not None:
                newGroupId = newGroup.find('id').text
                self.importGroupMembers(newGroupId, fullName, group.find(userGroupListObj), userGroupObj)
                count += 1
            existGroupId = None

        self.log.info("New groups created: %s" % count)
        self.log.info("importGroups finished successfully")

    def importGroupMembers(self, groupId, fullName, members, userGroupObj):
        if members is None or groupId is None or not isinstance(groupId, str):
            return
        elif groupId.find('group') != 0:
            return
        existMembers = self.getExistGroupMembers(groupId)
        cmd = 'addUserGroupMember'
        for m in members.iterfind(userGroupObj):
            userName = m.find('userName').text
            # check existing group membership from cached list
            if existMembers.xpath("//ActiveGroupMemeber[userName='%s']" % userName):
                self.log.debug("  skip '%s', user is already a member of '%s' group." % \
                               (userName, fullName))
                continue
            args = [groupId, userName]
            ApiMethods[cmd](self.session, cmd, args).export_xml()

            if self.session.error_code == 'ObjectAlreadyExistsFault':
                self.log.debug("  skip '%s', user is already a member of '%s' group." % \
                               (userName, fullName))
            elif self.session.error > 0 :
                self.log.error("  failed to add user '%s' to group '%s'" % (userName, fullName))
            else:
                self.log.debug("  add '%s' to group '%s'" % (userName, fullName))

