
#import os
import re
import sys
import CTFApp

from ImportModule import ApiMethods
from ImportModule import Commands
from SimpleLog import SimpleLog

# __revision__  = $LastChangedRevision$
# __Copyright__ = Copyright (c) 2007 - 2009, CollabNet, Inc. All rights are reserved.

commands = ['set', 'get', 'quit', 'status', 'server', 'setcontext',
            'unsetcontext', 'getcontext', 'which', 'deleteprojects']

def quit():
    sys.exit(0)          

class set(object):
    def __init__(self, session, args=None):
        self.session = session
        self.loglevel = self.session.loglevel
        self.args = args
        self.log = SimpleLog(self.loglevel)
        self.var = None
        self.val = None

    def usage(self, errmsg=None):
        if errmsg:
            self.log.error("%s" % errmsg)
        self.log.info("    Assign session variable to giving value:")
        self.log.info("    usage: set varname value\n")
        self.log.info("    Following variables can be set this way:\n")
        self.print_varlist(self.session.RW_attributes)

    def validate(self):
        if self.args is None:
            self.usage("    syntax error: missing arguments")
            return False
        elif len(self.args) == 2:
            self.var = self.args[0]
            self.val = self.args[1]
        elif len(self.args) == 1:
            if self.args[0] in ['xml', 'table'] :
                self.var = self.args[0]
            else:
                self.usage("    syntax error: missing arguments")
                return False
        elif not self.var in self.session.RW_attributes:
            self.usage("    unknown variable or the variable can't be set manually.")
            return False
        return True
           
    def run(self):
        if not self.validate():
            return
        if self.var is None:
            self.usage()
        elif self.var == 'loglevel':
            if self.val.upper() in self.log.loglevel:
                self.session.loglevel = self.val.upper()
            else:
                self.log.warn("unknown loglevel: %s" % self.val)
        elif self.var == 'soaptrace':
            if self.val is not None and self.val.lower() == 'true' or self.val.lower() == 'on':
                self.session.kw_soapopt = {'tracefile' : sys.stderr}
            elif self.val is not None and self.val.lower() == 'false' or self.val.lower() == 'off':
                self.session.kw_soapopt = {}
            elif self.val is not None and self.val != 'sys.stderr' and self.val != 'sys.stdout':
                try:
                    fh = open(self.val, 'a')
                    self.session.kw_soapopt = {'tracefile' : fh}
                except IOError, err:
                    self.log.error("can't open debugfile %s" % self.val)
                    self.log.error("with error message: %s" % err)
                    self.log.warn("fall back to using sys.stderr")
                    self.session.kw_soapopt = {'tracefile' : sys.stderr}
            self.session.soaptrace = self.val
                
        elif self.var == 'xml':
            self.session.format = 'xml'
        elif self.var == 'table':
            self.session.format = 'table'
        elif self.var in CTFApp.CONTEXT_NAMES:
            setcontext(self.session, self.args).run()
        else:
            self.session.__setattr__(self.var, self.val)

    def print_varlist(self, varlist):
        MAX_WIDTH = 72
        varlist.sort()
        num_cmds = len(varlist)
        max_len = 0
        for c in varlist:
            if len(c) > max_len:
                max_len = len(c)
        col_width = max_len + 2
        ncol = MAX_WIDTH / col_width
        nrow = num_cmds / ncol
        if num_cmds % ncol > 0 :
            nrow += 1

        for n in range(nrow):
            line = '    '
            for col in range(ncol):
                N = col * nrow + n
                if N < num_cmds:
                    line += varlist[N].ljust(col_width)
            print >> sys.stderr, line

class get(object):
    def __init__(self, session, args=None):
        self.session = session
        self.loglevel = self.session.loglevel
        self.log = SimpleLog(self.loglevel)
        self.args = args
        self.var = None

    def usage(self, errmsg=None):
        if errmsg:
            self.log.error("%s" % errmsg)
        self.log.info("Get session variable value:")
        self.log.info("    usage:  get varname")
        self.log.info("       if varname is 'all', then all variables are listed")

    def validate(self):
        if self.args and len(self.args) == 1:
            self.var = self.args[0]
        else:
            self.usage("syntax error")
            return False
        if self.var == 'all':
            return True
        elif not self.var in self.session.RW_attributes:
            self.usage("unknown variable")
            return False
        return True
           
    def run(self):
        if not self.validate():
            return
        if self.var == 'all':
            for k in  self.session.RW_attributes:
                self.log.info("    %s = %s" % (k.ljust(20),
                                                  self.session.__getattribute__(k)))
        else:
            self.log.info("    %s = %s" % (self.var,
                                              self.session.__getattribute__(self.var)))

class server(object):

    def __init__(self, session, args=None):
        self.session = session
        self.loglevel = self.session.loglevel
        self.log = SimpleLog(self.loglevel)
        self.args = args
        self.url = None

    def usage(self, errmsg=None):
        if errmsg:
            self.log.error("%s" % errmsg)
        self.log.info("Setting base URL to TeamForge server.")
        self.log.info("Usage:  server http://hostname:port")

    def validate(self, args):
        if args and len(args) == 1:
            self.url = args[0]
            if not self.url.startswith('http'):
                self.usage("format error, must provide a valid URL")
                return False
        else:
            self.usage("syntax error")
            return False
        return True
           
    def run(self):
        if not self.validate(self.args):
            return
        self.session.base_url = self.url
        self.session.set_app_url()

class which(object):
    def __init__(self, session, args=None):
        self.session = session
        self.loglevel = session.loglevel
        self.log = SimpleLog(self.loglevel)
        self.args = args
        self.cmd = None
      #  self.run()

    def usage(self, errmsg=None):
        if errmsg:
            self.log.error("%s" % errmsg)
        self.log.info("show module name for command")
        self.log.info("Usage: which <command name>")
    
    def validate(self):
        if len(self.args) == 1:
            self.cmd = self.args[0]
        else:
            self.usage("syntax error")
            return False
        return True

    def run(self):
        if not self.validate():
            return
        else:
            if self.cmd in Commands:
                self.log.info("%s => %s" % (self.cmd, Commands[self.cmd]))
            elif self.cmd in ApiMethods:
                self.log.info("%s => %s" % (self.cmd, ApiMethods[self.cmd]))


#class quit(object):
#    def __init__(self, session=None, args=None):
#        self.run()
#
#    def usage(self, errmsg=None):
#        print >> sys.stderr, "INFO : Exit the API tool normally."
#
#    def run(self):
#        sys.exit(0)

class status(object):
    def __init__(self, session, args=None):
        self.session = session
        self.loglevel = self.session.loglevel
        self.log = SimpleLog(self.loglevel)
        self.args = args

    def usage(self):
        self.log.info("print out login session status")

    def run(self):
        self.log.info( "Session status:")
        self.log.info( "       Base URL: %s" % self.session.base_url)
        self.log.info( "     user name : %s" % self.session.user)
        if self.session.session_id:
            self.log.info( "        session: logged in")
        else:
            self.log.info( "        session: None (not logged in)")

class setcontext(object):
   
    def __init__(self, session, args=None):
        self.args = args
        self.session = session
        self.session.error = 0
        self.loglevel = self.session.loglevel
        self.log = SimpleLog(self.loglevel)

    def usage(self, errmsg=None):
        if errmsg:
            self.log.error("%s" % errmsg)
        self.log.info ("""
     set session contexts:
     Usage: setcontext contextName value
       where contextName = project|page|user|group|tracker|wiki|planningFolder|documentFolder
     For example: 
         setcontext project proj1001
         setcontext page    page1001""")

    def validate(self):
        if self.args is None or len(self.args) < 2:
            self.usage("missing argument")
            return False
        return True 

    def run(self):
        if self.validate():
            ctxName = self.args[0]
            ctxVal  = self.args[1]
            objName = ctxName[0].upper() + ctxName[1:]
            cmdName = 'get' + objName + 'Data'

            if cmdName in ApiMethods:
                data = ApiMethods[cmdName](self.session, 
                                       cmdName,
                                       [ctxVal]).export_xml()
                if self.session.error > 0:
                    return
            else:
                self.log.warn("You can not use setcontext on this type of object")
                self.log.info("There is no such SOAP call: '%s'" % cmdName)
                self.session.error = 1
                return

            if data is not None:
                self.session.Ctx[ctxName] = getattr(CTFApp, objName)(data)
            else:
                self.usage("Context not set for %s '%s'" % (ctxName, ctxVal))

class getcontext(object):
    def __init__(self, session, args=None):
        self.session = session
        self.loglevel = self.session.loglevel
        self.log = SimpleLog(self.loglevel)
        self.args = args

    def usage(self, errmsg=None):
        if errmsg:
            self.log.error("%s" % errmsg)
        self.log.info("get current session contexts")
        self.log.info("Usage: getcontext")

    def run(self):
        self.log.info("Current session contexts:")
        for item in self.session.Ctx:
            xid = self.session.Ctx[item].id
            item_attr = dir(self.session.Ctx[item])
            if 'title' in item_attr:
                title = self.session.Ctx[item].title
            elif 'name' in item_attr:
                title = self.session.Ctx[item].name
            elif 'username' in item_attr:
                title = self.session.Ctx[item].username
            elif 'fullName' in item_attr:
                title = self.session.Ctx[item].fullName
            self.log.info("  %s context = %s (%s)" % (item.ljust(10), xid.ljust(10), title))

class unsetcontext(object):
    def __init__(self, session, args=None):
        self.session = session
        self.loglevel = self.session.loglevel
        self.log = SimpleLog(self.loglevel)
        self.args = args

    def usage(self, errmsg=None):
        if errmsg:
            self.log.error("%s" % errmsg)
        self.log.info("unset current session context")
        self.log.info("Usage: unsetcontext project|usre|group|role|page|tracker")

    def validate(self):
        if self.args is None or len(self.args) < 1:
            self.usage("missing argument")
            return False
        return True 

    def run(self):
        if not self.validate():
            return
        ctxName = self.args[0]
        if ctxName in self.session.Ctx:
            self.session.Ctx.pop(ctxName)

class deleteprojects(object):
    def __init__(self, session, args=None):
        self.session = session
        self.loglevel = self.session.loglevel
        self.log = SimpleLog(self.loglevel)
        self.args = args
        self.input = []

    def usage(self, errmsg=None):
        if errmsg:
            self.log.error("%s" % errmsg)
        self.log.info("delete multiple projects ")
        self.log.info('Usage: deleteprojects "proj1050, proj1100 - proj1120"')

    def validate(self):
        if self.args is None or len(self.args) < 1:
            self.usage("missing argument")
            return False
        for item in self.args[0].split(','):
            if item.find('-') > 0:
                start, end = item.split('-')
                m1 = re.match('proj(\d{4,})', start.strip())
                if m1 and m1.group(1) is not None:
                    r1 = int(m1.group(1))
                m2 = re.match('proj(\d{4,})', end.strip())
                if m2 and m2.group(1) is not None:
                    r2 = int(m2.group(1))
                for i in range(r1, r2+1):
                    self.input.append("proj"+str(i))
            else:
                self.input.append(item.strip())
        return True 

    def run(self):
        if not self.validate():
            return
        cmd = 'deleteProject'
        for projectId in self.input:
            ApiMethods[cmd](self.session, cmd, [projectId]).run()
            if self.session.error > 0:
                self.log.error("Can't delete project '%s'" % projectId)
            else:
                self.log.info("deleting project '%s'" % projectId)
