
# __revision__  = $LastChangedRevision$
# __Copyright__ = Copyright (c) 2007 - 2009, CollabNet, Inc. All rights reserved.

import os
import sys
from lxml import etree
from ZSI import FaultException

BASEDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
sys.path.append(os.path.join(BASEDIR, "TeamForge"))

from SimpleLog import SimpleLog
from Soap60.CollabNet import CollabNetSoapService_services as CollabNetServices
import DataUtils
import CTFApp

# bind to a dummy url in order to get all available services
Services = [m for m in dir(CollabNetServices.CollabNetSoapBindingSOAP('')) \
               if not (m.startswith('_') or m.startswith('binding'))]
SoapApp = 'CollabNet'

class CollabNet(object):
    def __init__(self, session, cmd, args=None):
        self.session = session
        self.session.error = 0
        self.session.error_code = None
        self.session.errmsg = None
        self.isCTF53 = None
        self.loglevel = self.session.loglevel
        self.cmd = cmd
        self.args = args
        self.results = None
        self.log = SimpleLog(self.loglevel, self.session.logfile)

        if self.session.base_url is None:
            self.log.error("TeamForge server is not specified.")
            return
        else:
            self.port = self.get_service_binding()
        if self.port is None:
            self.log.error("SoapApp URL is not set for '%s'" % SoapApp)
            return

        self.request = self.get_request()
        self.params, self.ptypes = DataUtils.get_params_types(self.request)
        self.tagname = DataUtils.get_tagname(self.cmd)

        if self.loglevel == 'DEBUG5':
            self.log.debug5("")
            self.log.debug5("request parameters for cmd '%s':" % self.cmd)
            for idx in range(len(self.params)):
                self.log.debug5("    request.%s (%s) " % (self.params[idx],
                                                   self.ptypes[idx]))
        if not self.session.help:
            self.action()

    def get_service_binding(self):
        url = self.session.get_app_url(SoapApp)
        if url:
            port = CollabNetServices.CollabNetSoapBindingSOAP(url, **self.session.kw_soapopt)
        else:
            return None
        return port

    def get_request(self):
        req_name = self.cmd + "Request"
        return getattr(CollabNetServices, req_name)()

    def usage(self, errmsg=None):
        if errmsg:
            self.log.error("%s" % errmsg)
        if self.cmd.startswith('login'):
            self.log.info("Usage: %s %s" % (self.cmd, ' '.join(self.params)))
        else:
            self.log.info("Usage: %s %s" % (self.cmd, ' '.join(self.params[1:])))

    def validate(self):
        dont_use_sessionId = ['getApiVersion', 'login', 'loginAnonymous',
                              'loginWithToken', 'invalidateEmailTemplate',
                              'invalidateResourceBundleCache', 'logoff'] 
        need_projectId = ['addProjectMember', 'deleteProject', 'getProjectAccessLevel',
                          'getProjectData', 'getProjectDiskUsage', 'getProjectGroupList',
                          'getProjectMemberList', 'getProjectQuota', 'hasPermission',
                          'listProjectAdmins', 'removeProjectMember', 'setProjectAccessLevel',
                          'setProjectQuota']

        if not self.cmd in dont_use_sessionId and self.session.session_id is None:
        #if not (self.cmd in dont_use_sessionId or self.session.session_id):
            self.log.error("active session not found, please login first")
            self.log.info("cmd = '%s'" % self.cmd)
            return False

        # session_id is always the first argument for rest of methods
        if self.cmd not in dont_use_sessionId and self.session.session_id:
            self.args.insert(0, self.session.session_id)

        # if project context is set and user didnot provide projectId, 
        # use projectId from project context
        if self.session.Ctx.has_key('project') and  \
           self.session.Ctx['project'] is not None and \
           self.cmd in need_projectId and \
           len(self.args) == len(self.params) - 1: 
            self.args.insert(1, self.session.Ctx['project'].id)


        # if filter is not given, append an empty one
        if self.cmd in ['getUserList', 'getUserGroupList'] and \
           len(self.args) == len(self.params) - 1: 
            self.args.append('')
        num_args = len(self.args)

        if self.cmd == 'logoff':
            if self.args == []:
                self.args = []
                self.args.append(self.session.user)
                self.args.append(self.session.session_id)
            elif len(self.args) == 1:
                self.args.append(self.session.session_id)

        if self.cmd == "getProjectList":
            # flag indicating whether to fetch the hierarchy path for each project
            self.args.append('True')

        if len(self.args) < len(self.params) :
            self.usage("syntax error: missing argument")
            return False
        elif len(self.args) > len(self.params) :
            self.usage("syntax error: too many argument")
            return False
        else:
            for idx in range(len(self.params)):
                method = "set_element_" + self.params[idx]
                if self.ptypes[idx] == 'boolean':
                    val = DataUtils.to_boolean(self.args[idx])
                elif self.ptypes[idx] == 'int':
                    val = DataUtils.to_int(self.args[idx])
                elif self.ptypes[idx] == 'long':
                    val = DataUtils.to_long(self.args[idx])
                elif self.ptypes[idx] == 'string':
                    val = DataUtils.to_string(self.args[idx])
                elif self.ptypes[idx] == 'SoapFilter':
                    #val = self.creat_Filter(self.args[idx])
                    val = None
                elif self.ptypes[idx] == 'ArrayOf_xsd_string':
                    val = DataUtils.to_array(self.args[idx])
                else: 
                    val = self.args[idx]
    
                if val is not None:
                    getattr(self.request, method)(val)
    
                if self.loglevel == 'DEBUG5':
                    getter = method.replace('set', 'get')
                    self.log.debug5("    request.%s => %s" % \
                                    (getter, getattr(self.request, getter)()))
        return True

    def action(self):
        if not self.validate():
            return
        try:
            response = getattr(self.port, self.cmd)(self.request)
        except FaultException, fe:
            fc = fe.fault.code[1]
            fs = fe.fault.string
            if fc in ['LoginFault', 'ObjectAlreadyExistsFault']:
                pass
            else:
                if not self.cmd.startswith('login'):
                    self.log.debug5("cmd failed: '%s %s'" % (self.cmd, self.args))
                self.log.error("FaultException, error code    = %s" % fc)
                self.log.error("FaultException, error message = %s" % fs)
            self.session.error = 1
            self.session.error_code = fc
            return
        except TypeError, te:
            major, minor = DataUtils.get_PythonVersion()[0:2]
            if major > 2 or major == 2 and minor > 4:
                fs = repr(te)
            else:
                fs = te
            if not self.cmd.startswith('login'):
                self.log.debug5("cmd failed: '%s %s'" % (self.cmd, self.args))
            self.log.error("TypeError, error message = %s" % fs)
            self.session.error = 1
            return
        except Exception, ee:
            major, minor = DataUtils.get_PythonVersion()[0:2]
            if major > 2 or major == 2 and minor > 4:
                fs = repr(ee)
            else:
                fs = ee
            if not self.cmd.startswith('login'):
                self.log.debug5("cmd failed: '%s %s'" % (self.cmd, self.args))
            self.log.error("Exception, error message = %s" % fs)
            self.session.error = 1
            self.session.error_code = fs
            import re
            mm = re.match(".*?'(.*?)'.*", '%s' % fs)
            if mm and mm.group(1):
                self.session.errmsg = "ERROR: %s" % mm.group(1)
            return
        if response:
            getReturn = 'get_element_' + self.cmd + 'Return'
            if getReturn in dir(response):
                try:
                    self.results = getattr(response, getReturn)()
                except Exception, ee:
                    self.log.error("error message: %s" % repr(ee))

#        if self.cmd.startswith('login'):
#            self.session.Version = CollabNet(self.session, 'getVersion', []).export_data()

    def run(self):
        if self.cmd == 'logoff':
            self.session.session_id = None
            self.session.Ctx = {}
            self.session.indices = {}
        elif self.results is None:
            return
        elif self.cmd.endswith('Version'):
            print >> sys.stderr, "%s" % self.results.encode('utf-8')
        elif self.cmd.startswith('login'):
            self.session.session_id = self.results
            CMD = 'getVersion'
            self.session.Version = CollabNet(self.session, CMD, []).export_data()
            CMD = 'getApiVersion'
            self.session.ApiVersion = CollabNet(self.session, CMD, []).export_data()
            self.session.user = self.args[0]
            self.isCTF53 = DataUtils.isSameOrNewer(self.session.min_version, self.session.Version)
            # reset session context
            self.session.Ctx = {}
            return
        elif isinstance(self.results, bool):
            if self.results:
                print >> sys.stderr, "True"
            else:
                print >> sys.stderr, "False"
            return
        elif self.cmd.startswith('create'):
            if self.cmd.startswith("createProject"):
                project_data = self.export_xml()
                if project_data is not None:
                    self.session.Ctx['project'] = CTFApp.Project(project_data)
            elif self.cmd == "createUser":
                user_data = self.export_xml()
                if user_data is not None:
                    self.session.Ctx['user'] = CTFApp.User(user_data)
            elif self.cmd == "createGroup":
                group_data = self.export_xml()
                if group_data is not None:
                    self.session.Ctx['group'] = CTFApp.Group(group_data)
        else:
            xml = self.export_xml()
            if xml is None:
                return
            elif self.session.format == 'xml':
                print  >> sys.stderr, etree.tostring(xml, encoding='UTF-8', pretty_print=True)
            elif self.session.format == 'table':
                DataUtils.export_table(xml)

    def export_xml(self):
        return DataUtils.export_xml(self.results, root_name=self.tagname)

    def export_dict(self):
        return DataUtils.export_dict(self.results)

    def export_data(self):
        return self.results

#    def create_Filter(self, val):
#        # for now, filter is ignored, so just return a dummy
#        sf = self.request.new_filter()
#        sf.name = ''
#        sf.value = ''
#        return sf
