
# __revision__  = $LastChangedRevision$
# __Copyright__ = Copyright (c) 2007 - 2009, CollabNet, Inc. All rights reserved.

import os
import re
import sys
import tarfile
import StringIO
import time

from lxml import etree

def get_PythonVersion():
    """Returns python version as (major, minor, patch), eg (2, 4, 3)"""
    pv =  sys.version.split()[0].split('.')
    return [int(m) for m in pv]
    #return pv

def getPythonVersion():
    """Returns python version as (major, minor, patch), eg (2, 4, 3)"""
    pv =  sys.version.split()[0]
    return pv

def isSameOrNewer(min_version=None, version=None):
    """Compares if version is the same or newer than min_version."""
    if version is None:
        return False
    elif min_version is None and version is not None:
        return True

    vm_list = min_version.split('.')
    v_list = version.split('.')
    n = min(len(v_list), len(vm_list))
    for i in range(0, n):
        if v_list[i] > vm_list[i]:
            return True
        elif v_list[i] < vm_list[i]:
            return False
    return True

def normalizer(s):
    """This method normalizes input string by performing following transforms:
       1) replace non-alphanumeric characters with space.
       2) replace multiple spaces with a single space.
    """
    a = ''
    for x in s:
        if x.isalnum() or x.isspace():
            a += x
        else:
            a += ' '
    b = ' '.join([x for x in a.split(' ') if x])
    return b
    
def attr_getter(soap_object):
    return [x for x in dir(soap_object) if x.startswith('get_element_')]

def attr_setter(soap_object):
    return [x for x in dir(soap_object) if x.startswith('set_element_')]

#TODO: this is for getting xxxReturn object from Soap responses.
#      If we cache return objects, then this op should be moved into
#      soap commands so that we only have to extract return from response once.
def get_soap_payload(soap_object):
    get_method = attr_getter(soap_object)
    if len(get_method) == 1 and get_method[0].endswith('Return'):
        return getattr(soap_object, get_method[0])()
    return soap_object

def export_SoapDO(soap_do, xml):
    if soap_do == None:
        return
    for m in attr_getter(soap_do):
        name = m.replace('get_element_', '')
        val = getattr(soap_do, m)()
        if val is None or val == '' or val == [] :
            # subelements without value
            etree.SubElement(xml, name)
        elif name == 'assignments' or \
            name == 'fieldValues' or \
            name == 'flexFields'  or \
            name == 'settings' :
            xml.append(export_xml(val, root_name=name))
        else:
            etree.SubElement(xml, name).text = unicode(val)
     
def export_dict(soap_object):
    retval = {}
    if not soap_object:
        return retval
    
    for m in attr_getter(soap_object):
        name = m.replace('get_element_', '')
        val = getattr(soap_object, m)()
        retval[name] = val
    return retval

def export_SoapList(soap_do, xml):
    data_rows = soap_do.get_element_dataRows()
    if len(data_rows) > 0:
        tagname = xml.tag.rstrip('s')
        for row in data_rows:
            xml.append(export_xml(row, root_name=tagname))


def export_table(xml):
    FORMAT = "INFO :  %s  %s"
    entries = []
    max_size = 6
    t1 = t2 = ''
    if xml.text:
        print >> sys.stderr, ("%s" % xml.text.encode('utf-8'))
    elif xml.find('id') is None :
        # for soap answers don't contain 'id' 
        if xml.tag in ['ActiveGroupMember']:
            t1, t2, xid, title = export_table_row(item)
        else:    
            for item in xml:
                t1, t2, xid, title = export_table_row(item)
                if xid is not None and title is not None:
                    entries.append((xid, title)) 
                    max_size = max(len(xid), len(t1), max_size)
                else:
                    # result has no 'id' node
                    print "FIXME: export_table(): ", t1, t2, xid, title
    else:
        t1, t2, xid, title = export_table_row(xml)
        if xid is not None and title is not None:
            entries.append((xid, title)) 
            max_size = max(len(xid), len(t1), max_size)
    if entries:
        print >> sys.stderr, FORMAT % (t1.ljust(max_size), t2.encode('utf-8'))
        print >> sys.stderr, "INFO :  -----------------------------------------------------"
        for e in entries:
            print >> sys.stderr, FORMAT % (e[0].ljust(max_size), e[1].encode('utf-8'))

def export_table_row(xml):
    id_name = 'id'
    title_name = 'title'
    xid = ''
    title = ''
    
    if xml is not None:
        if xml.find('id') is not None and xml.find('title') is not None:
            xid = xml.find('id').text
            title = xml.find('title').text
        elif xml.tag in ['Attachment', 'PageImages']:
            xid = xml.find('attachmentId').text
            title = xml.find('fileName').text
            title_name = 'fileName'
        elif xml.tag == 'Group':
            # getGroupList does not return id, only getGroupList2 does
            gid_node =  xml.find('id')
            if gid_node is None:
                title_name = 'fullName'
                title = xml.find('fullName').text
                return id_name, title_name, '-', title
            else:
                xid = gid_node.text
                title = xml.find('fullName').text
                title_name = 'fullName'
        elif xml.tag == 'ChildDependency':
            xid = xml.find('targetId').text
            title = xml.find('targetTitle').text
            id_name = 'targetId'
            title_name = 'targetTitle'
        elif xml.tag in ['Field']:
            xid = xml.find('id').text
            title = xml.find('name').text
            title_name = 'fieldName'
        elif xml.tag in ['PlanningStatusValue']:
            xid = xml.find('id').text
            title = xml.find('value').text
            title_name = 'value'
        elif xml.tag in ['ProjectAdmin', 'ProjectMember', 'User', 'ActiveGroupMember']:
            id_node = xml.find('id')
            if id_node is None:
                xid = xml.find('userName').text
                title = xml.find('fullName').text
                id_name = 'userName'
                title_name = 'fullName'
            else:
                xid = id_node.text
                title = '%s (%s)' % (xml.find('fullName').text,
                                     xml.find('username').text)
                title_name = 'fullName (username)'
        elif xml.tag in ['ProjectGroup']:
            xid = xml.find('fullName').text
            title = xml.find('description').text
            title_name = 'Description'
        elif xml.tag in ['WikiPage']:
            xid = xml.find('id').text
            title = xml.find('wikiPageName').text
        else:
            print "export_table_row(): WARNING unprocessed xml node: tag name: %s" % xml.tag
            print "  unprocessed xml node:"
            print etree.tostring(xml, pretty_print=True, encoding='utf-8')
            return

        #print >> sys.stderr, FORMAT % (xid.ljust(20), title.encode('utf-8'))
        if title is None:
            title = ''
        if xid is None:
            xid = ''
        return id_name, title_name, xid, title
    
def export_xml(soap_object, root_name=None):
    xml_root = etree.Element(root_name)
    retval = None
    if not soap_object:
        return
    elif isinstance(soap_object, int):
        xml_root.text = str(soap_object)
        return xml_root
    elif isinstance(soap_object, unicode):
        xml_root.text = str(soap_object)
        return xml_root
       
    retval = get_soap_payload(soap_object)

    if hasattr(retval, 'DataRows'):
        # return val is a list of SoapObjects
        export_SoapList(retval, xml_root)       
    elif isinstance(retval, list):
        # return val is a list of SoapDO
        tagname = xml_root.tag.rstrip('s')
        for entry in retval:
            xml_root.append(export_xml(entry, root_name=tagname))
    else:
        # return val is a single SoapDO
        export_SoapDO(retval, xml_root)
    return xml_root

def get_tagname(cmd):
    name = cmd.replace('2', '')
    p1 = r'(get|list|find)(.*)(List|List2|s$|s2$)'
    m = re.match(p1, cmd)
    if m and m.group(2) != None:
        name = m.group(2)
        if name.endswith('2'):
            name = name[0:-1]
    if m and m.group(3) != None:
        suffix = 's'
    else:
        suffix = ''

    p2 = r'(get)(.*)Data'
    m2 = re.match(p2, cmd)
    if m2 and m2.group(2) != None:
        name = m2.group(2)
        if name.endswith('2'):
            name = name[0:-1]
    full = name + suffix
#    if full.endswith('ys'):
#        full = full[0:-2] + 'ies'
    return full

def to_array(value):
    if value is None or value == '':
        return None
    elif isinstance(value, list) and value:
        return value
    elif isinstance(value, str):
        if value.lower() == 'nil' or value.lower() == 'none':
            return None
        l = eval(value)
        if isinstance(l, list) and l:
            return l
        else:
            return value.split()

def to_boolean(value):
    if isinstance(value, bool):
        return value
    elif isinstance(value, str):
        v = value.lower()  
        if v == 'true' or v == 'yes':
            return True
        else:
            return False
    else:
        return False

def to_int(value):
    if value:
        return int(value)
    else:
        return 0

def to_long(value):
    if value:
        return long(value)
    else:
        return 0.0

def to_string(value):
    if value:
        if value.lower() == 'nil' or value.lower() == 'none':
            return None
        else:
            return value.strip('"|\'')
    else:
        return ''

def to_dateTime(value):
    y = m = d = H = M = S = ms = 0
    if isinstance(value, tuple):
        return value
    #try to see if the str can be eval'ed into a tuple:
    try:
        val = eval(value)
        if isinstance(val, tuple):
            return val
    except:
        pass

    # value must be a str
    if value is None:
        return None
    elif isinstance(value, str):
        if value == '' or value.lower() == 'none' \
           or value.lower() == 'nil':
            return None

    if value.find('(') == 0:
        l = value.strip('\(\)').replace(',', ' ').split()
        dateTime = [int(a) for a in l]
        if dateTime[6] == 0:
            dateTime[6] = 1
        return dateTime
    m = re.match('(\d{4})[/|-](\d{1,2})[/|-](\d{1,2})([ |:](\d{1,2}):(\d{1,2}):(\d{1,2}))?', value)
    if m:
        value = value.replace('/', ' ')
        value = value.replace(':', ' ')
        value = value.replace('-', ' ')
        if m.group(4) is None:
            y, m, d =  value.split(' ')
        else :
            y, m, d, H, M, S =  value.split(' ')
        dateTime = (int(y), int(m), int(d), int(H), int(M), int(S), 0, 0, 0)
        return dateTime
    m = re.match('(\d{1,2})[/|-](\d{1,2})[/|-](\d{4})([ |:](\d{1,2}):(\d{1,2}):(\d{1,2}))?', value)
    if m:
        value = value.replace('/', ' ')
        value = value.replace(':', ' ')
        value = value.replace('-', ' ')
        if m.group(4) is None:
            m, d, y =  value.split(' ')
        else :
            m, d, y, H, M, S =  value.split(' ')
        dateTime = (int(y), int(m), int(d), int(H), int(M), int(S), 0, 0, 0)
        return dateTime

def get_params_types(request):
    """Return parameters and their class name of a request as lists"""
    ofwhat = request.typecode.ofwhat
    params = [ofwhat[idx].pname for idx in range(len(ofwhat))]
    ptypes = [ofwhat[idx].type[1] for idx in range(len(ofwhat))]
    return params, ptypes

def find_children(data):
    rootId = None
    node_dict = {}
    children = {}
    for node in data:
        parentId = node.find('parentFolderId').text
        xid = node.find('id').text
        node_dict[xid] = node
        if not parentId in children:
            children[parentId] = []
        children[parentId].append(xid)
    for key in children:
        if not key in node_dict:
            rootId = key
            break
    return (rootId, children, node_dict)

def print_treeview(rootId, mapping, nodes, indent):
    indent = indent + '   '
    if not rootId in mapping:
        return
    for child in mapping[rootId]:
        text = nodes[child].find('title').text.encode('utf-8')
        print >> sys.stderr, "INFO : %s|_ %s (%s)" % (indent, text, child)
        print_treeview(child, mapping, nodes, indent)

def create_archive(filename, ftype=None):
    """returns a file object """
    if os.path.exists(filename):
        print >> sys.stderr, "ERROR: File '%s' exists, either remove it or use another name."  % filename
        return
    if ftype == "tgz":
        return tarfile.open(filename, mode='w:gz')
    elif ftype == "tar":
        return tarfile.open(filename, mode='w')
    elif ftype is None:
        return open(filename, 'w')
      

#def write_data(tar, archive_name, xml_data):
#    #self.log.info("Startig to write data to archive file ...")
#    retval = True
#    if xml_data is not None:
#        fobj = StringIO.StringIO(etree.tostring(xml_data,
#                                                encoding='UTF-8',
#                                                pretty_print=True,
#                                                xml_declaration=True))
#        tarinfo = tarfile.TarInfo(name=archive_name)
#        tarinfo.size = len(fobj.buf)
#        tarinfo.uname = 'nobody'
#        tarinfo.gname = 'nobody'
#        tarinfo.mtime = time.time()         # seconds from Epoc
#        try:
#            tar.addfile(tarinfo=tarinfo, fileobj=fobj)
#        except Exception, e:
#            print >> sys.stderr, "failed to save data into archive"
#            print >> sys.stderr, "exception %s" % e
#            retVal = False
#    return retval

