
# __revision__  = $LastChangedRevision$
# __Copyright__ = Copyright (c) 2007 - 2009, CollabNet, Inc. All rights reserved.

import os
import sys
from lxml import etree
from ZSI import FaultException

BASEDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
sys.path.append(os.path.join(BASEDIR, "TeamForge"))

from Soap60.DocumentApp import DocumentAppSoapService_services as DocServices
from SimpleLog import SimpleLog
import DataUtils

# bind to a dummy url in order to get all available services
Services = [ m for m in dir(DocServices.DocumentAppSoapBindingSOAP('')) \
                  if not (m.startswith('_') or m == 'binding')]
SoapApp = 'DocumentApp'

class DocumentApp(object):
    def __init__(self, session, cmd, args=None):
        self.session = session
        self.session.error = 0
        self.loglevel = self.session.loglevel
        self.cmd = cmd
        self.args = args
        #self.mc = self.session.mc
        self.results = None
        self.log = SimpleLog(self.loglevel)

        if self.session.base_url is None:
            self.log.error("TeamForge server is not specified.")
            self.session.error = 1
            return
        else:
            self.port = self.get_service_binding()
        if self.port is None:
            self.log.error("SoapApp URL is not set for '%s'" % SoapApp)
            self.session.error = 1
            return

        self.request = self.get_request()
        self.params, self.ptypes = DataUtils.get_params_types(self.request)
        self.tagname = DataUtils.get_tagname(self.cmd)

        if self.loglevel == "DEBUG5":
            self.log.debug5("")
            self.log.debug5("request parameters for cmd '%s':" % self.cmd)
            for idx in range(len(self.params)):
                self.log.debug5("    request.%s (%s) " % (self.params[idx],
                                                   self.ptypes[idx]))

        if not self.session.help:
            self.action()

    def get_service_binding(self):
        url = self.session.get_app_url(SoapApp)
        if url:
            port = DocServices.DocumentAppSoapBindingSOAP(url, **self.session.kw_soapopt)
        else:
            return None
        return port

    def get_request(self):
        req_name = self.cmd + "Request"
        return getattr(DocServices, req_name)()

    def usage(self, errmsg=None):
        if errmsg:
            self.log.error("%s" % errmsg)
        self.log.info("Usage: %s %s" % (self.cmd, ' '.join(self.params[1:])))

    def validate(self):
        # session_id is always the first argument for all methods
        if not self.session.session_id:
            self.log.error("active session not found, please login first")
            return False
        else:
            self.args.insert(0, self.session.session_id)

        if len(self.args) < len(self.params) :
            self.usage("syntax error: missing argument")
            return False
        elif len(self.args) > len(self.params) :
            self.usage("syntax error: too many argument")
            return False
        else:
            for idx in range(len(self.params)):
                method = "set_element_" + self.params[idx]
                if self.ptypes[idx] == 'boolean':
                    val = DataUtils.to_boolean(self.args[idx])
                elif self.ptypes[idx] == 'int':
                    val = DataUtils.to_int(self.args[idx])
                elif self.ptypes[idx] == 'long':
                    val = DataUtils.to_long(self.args[idx])
                elif self.ptypes[idx] == 'string':
                    val = DataUtils.to_string(self.args[idx])
                elif self.ptypes[idx] == 'ArrayOf_xsd_string':
                    val = DataUtils.to_array(self.args[idx])
                elif self.ptypes[idx] == 'ArrayOf_tns1_SoapFilter':
                    #val = self.create_filter(self.args[idx])
                    val = None
                else:
                    val = self.args[idx]
                if val is not None:
                    getattr(self.request, method)(val)

                if self.loglevel >= self.log.loglevel['DEBUG5']:
                    getter = method.replace('set', 'get')
                    self.log.debug5("     request.%s => %s" % \
                          (getter, getattr(self.request, getter)()))
        return True

    def action(self):
        if not self.validate():
            return
        try:
            response = getattr(self.port, self.cmd)(self.request)
        except FaultException, fe:
            fc = fe.fault.code[1]
            fs = fe.fault.string
            self.log.debug5("cmd failed: '%s %s'" % (self.cmd, self.args))
            self.log.error("error code    = %s" % fc)
            self.log.error("error message = %s" % fs)
            self.session.error = 1
            return
        except TypeError, te:
            major, minor = DataUtils.get_PythonVersion()[0:2]
            if major > 2 or major == 2 and minor > 4:
                fs = repr(te)
            else:
                fs = te
            self.log.debug5("cmd failed: '%s %s'" % (self.cmd, self.args))
            self.log.error("TypeError, error message = %s" % fs)
            self.session.error = 1
            return

        if response:
            getReturn = 'get_element_' + self.cmd + 'Return'
            if getReturn in dir(response):
                try:
                    self.results = getattr(response, getReturn)()
                except Exception, ee:
                    self.log.error("error message: %s" % repr(ee))

    def run(self):
        if self.results is None:
            return
        elif isinstance(self.results, bool):
            if self.results:
                print >> sys.stderr, "True"
            else:
                print >> sys.stderr, "False"
            return
        elif self.cmd.startswith('create'):
            pass
        else:
            xml = self.export_xml()
            if xml is None:
                return
            elif self.session.format == 'xml':
                print  >> sys.stderr, etree.tostring(xml, encoding='UTF-8', pretty_print=True)
            elif self.session.format == 'table':
                DataUtils.export_table(xml)

    def export_xml(self):
        return DataUtils.export_xml(self.results, root_name=self.tagname)

    def export_dict(self):
        return DataUtils.export_dict(self.results)

    def export_data(self):
        return self.results

    def create_filter(self, val):
        """Filter is not used at the moment"""
        return
