
from lxml import etree

from ImportModule import ApiMethods
from SimpleLog import SimpleLog

import os
import re
import sys
import tarfile
import StringIO
import time
import DataUtils
import CTFApp

from urllib import urlencode
from cookielib import CookieJar
import urllib2

# __revision__  = $LastChangedRevision$
# __Copyright__ = Copyright (c) 2007 - 2009, CollabNet, Inc. All rights reserved.

# have to reload sys, in order to use this damn method!
reload(sys)
sys.setdefaultencoding("utf-8")

commands = ['export']

class HTTPDownloader(object):
    """This class uses form-based authentication to login to either HTTP or HTTPS
       site to download page pointed to by an url"""

    def __init__(self, base_url, user, password, level='INFO'):
        self.username = user
        self.password = password
        if base_url.endswith('/'):
            self.base_url = base_url.rstrip('/')
        else:
            self.base_url = base_url
        self.log = SimpleLog(level)
        cj = CookieJar()
        opener = urllib2.build_opener(urllib2.HTTPCookieProcessor(cj),
                                      urllib2.HTTPHandler(),
                                      urllib2.HTTPSHandler())
        urllib2.install_opener(opener)

        # to debug request/response:
        #   urllib2.HTTPHandler(debuglevel=1),
        #   urllib2.HTTPSHandler(debuglevel=1))

        login_page = self.base_url + '/sf/sfmain/do/login'
        params = urlencode({'sfsubmit' : 'submit',
                            'username' : self.username,
                            'password' : self.password
                           })
        request = urllib2.Request(login_page, params)
        try:
            response = urllib2.urlopen(request)
        except urllib2.URLError, ue:
            self.log.error("  URLError:")
            self.log.error("  errorno = '%s'" % ue.errno)
            self.log.error("  message = '%s'" % ue.strerror)
            sys.exit(1)

        if 'Workspace' not in response.read():
            self.log.error("Failed to login to '%s' page" % login_page)
            sys.exit(1)

    def getFile(self, page=''):
        """Return content at URL"""
        url = self.base_url + page
        self.log.debug("    url: %s" % url)
        req = urllib2.Request(url)
        try:
            resp =  urllib2.urlopen(req)
        except urllib2.HTTPError, he:
            self.log.error("HTTP error, url = '%s'" % he.url)
            self.log.error("        message = '%s'" % he.msg)
            self.log.error("    return code = '%s'" % he.code)
            return ''
        return resp.read()

class export(object):
    def __init__(self, session, args=None):
        self.session = session
        self.args = args
        self.projectId = None
        self.format = self.session.format
        self.data = None
        self.min_version = '5.3.0.0'      # This is static data, don't change it when branching
        self.tar = None
        self.base_dir = "ProjectFiles"
        self.log = SimpleLog(self.session.loglevel)
        self.HTTPDownloader = HTTPDownloader(self.session.base_url,
                                             self.session.user,
                                             self.session.passwd,
                                             self.session.loglevel)

    def create_archive(self, filename):
        if os.path.exists(filename):
            self.log.error("File '%s' already exists. Please remove it first."  % filename)
            sys.exit(1)
        self.tar = tarfile.open(self.session.outfile, 'w:gz')

    def usage(self, errmsg=None):
        if errmsg:
            self.log.error(errmsg)
        self.log.info("Exporting project data as a gzipped tar file (used with --out option):")
        self.log.info("Usage: export projectId")

    def validate(self):
        self.log.verbose("Validating user input")
        if not self.session.session_id:
            self.log.warn("no active session, please log in first.")
            return False

        if self.args and len(self.args) == 1:
            self.projectId = self.args[0]
            match = re.match('(proj\d{4,})', self.projectId)
            if match:
                # we have a projectId
                project_data = ApiMethods['getProjectData'](self.session,
                                                            'getProjectData',
                                                            [self.projectId]).export_xml()
                if self.session.error > 0:
                    self.log.error("failed to verify project '%s'" % self.projectId)
                    sys.exit(1)
            else:
                self.log.error("'%s' is not a valid projectId." % self.projectId)
                return
            if project_data is not None:
                self.session.Ctx['project'] = CTFApp.Project(project_data)
            else:
                self.log.error("project %s not found." % self.projectId)
                return False
        else:
            self.usage("syntax error, missing argument")
            return False

        self.log.debug("  Checking access to archive file.")
        if self.session.outfile is not None:
            self.create_archive(self.session.outfile)
        else:
            self.log.error("Please use '--out' option to provide an output filename")
            self.log.error("Or use 'set outfile <filename>' command in an interactive shell")
            return False

        self.log.verbose("Input validation complete.")
        return True
        
    def run(self):
        if not self.validate():
            return
        self.log.info("Startig to export data ...")
        self.data = etree.Element("TeamForge-Project")
        self.data.set('Version', self.session.Version)
        self.data.set('ApiVersion', self.session.ApiVersion)
        self.data.append(self.export_ProjectData(self.projectId))
        self.data.append(self.export_ProjectAccessLevel(self.projectId))
        self.data.append(self.export_TrackerApp(self.projectId))
        self.data.append(self.export_PlanningApp(self.projectId))
        self.data.append(self.export_WikiApp(self.projectId))
        self.data.append(self.export_DocumentApp(self.projectId))
        self.data.append(self.export_PageApp(self.projectId))
        self.data.append(self.export_RbacApp(self.projectId))
        self.data.append(self.export_FrsApp(self.projectId))
        self.write_data()
        self.log.info("Exporting all project data completed successfully.")

    def write_data(self):
        self.log.info("Startig to write data to archive file ...")
        if self.data is not None:
            fobj = StringIO.StringIO(etree.tostring(self.data, 
                                                    encoding='UTF-8',
                                                    pretty_print=True, 
                                                    xml_declaration=True))
            tarinfo = tarfile.TarInfo(name="project.xml")
            tarinfo.size = len(fobj.buf)
            tarinfo.uname = 'nobody'
            tarinfo.gname = 'nobody'
            tarinfo.mtime = time.time()         # seconds from Epoc
            try:
                self.tar.addfile(tarinfo=tarinfo, fileobj=fobj)
            except Exception, e:
                self.log.error("failed to save project data into archive")
                self.log.error("exception %s" % e)
                sys.exit(1)
        else:
            self.log.warn("project data is emptry, nothign to export")
        self.tar.close()
        self.log.info("Writing data complete, closing archive file.")

    def get_apiVersion(self):
        apiVersion = ApiMethods['getApiVersion'](self.session, 
                          'getApiVersion', []).export_xml()
        return apiVersion.text

    def export_ProjectData(self, projectId):
        self.log.info("Exporting project definition.")
        return ApiMethods['getProjectData'](self.session, 
                          'getProjectData',
                          [projectId]).export_xml()

    def export_ProjectAccessLevel(self, projectId):
        self.log.info("Exporting project access level.")
        al = ApiMethods['getProjectAccessLevel'](self.session, 
                          'getProjectAccessLevel',
                          [projectId]).export_xml()
        al.tag = 'AccessLevel'
        self.log.info("")
        return al

    def export_ProjectRoles(self):
        pass

    def add_to_archive(self, path, content):
        self.log.debug("    adding '%s' to archive file" % os.path.basename(path))
        tarinfo = tarfile.TarInfo(name=path)
        tarinfo.size = len(content.buf)
        tarinfo.uname = 'nobody'
        tarinfo.gname = 'nobody'
        tarinfo.mtime = time.time()         # seconds from Epoc
        self.tar.addfile(tarinfo=tarinfo, fileobj=content)
        
    def export_PageApp(self, projectId):
        self.log.info("Exporting project pages.")
        pageApp = etree.Element('PageApp')
        pages = etree.SubElement(pageApp, 'Pages')
        pageList = []
        if projectId:
            pageList = ApiMethods['getPageList'](self.session, 
                                                'getPageList',
                                                [projectId]).export_xml()
            PageAppId = None
            for page in pageList:
                pageId = page.find('id').text
                title = page.find('title').text
                page_data = ApiMethods['getPageData'](self.session,
                                                 'getPageData',
                                                 [pageId]).export_xml()
                if self.session.error > 0:
                    self.log.error("Can't export page %s" % pageId)
                    sys.exit(1)
                else:
                    self.log.verbose("  exporting page '%s'" % title)
                self.export_PageComponents(page_data)
                pages.append(page_data)
                if PageAppId is None:
                    if page.find('parentFolderId').text.startswith('PageApp'):
                        PageAppId = page.find('parentFolderId').text
        else:
            self.log.error("must provide a projectId as argument.")

        # export pageApp attachments
        pageApp.append(self.export_Attachments(PageAppId))

        self.log.info("Exporting project pages completed")
        self.log.info("")
        return pageApp

    def export_Attachments(self, objectId, Object=None):
        if objectId is not None:
            cmd = 'listAttachments'
            attachments = ApiMethods[cmd](self.session, cmd, [objectId]).export_xml()
            if self.session.error > 0:
                self.log.error("Can't export attachments for object '%s'" % objectId)
                sys.exit(1)
            else:
                self.log.debug("    processing attachments for object '%s'" % objectId)
        else:
            self.log.error("Must provide an objectId.")
            sys.exit(1)
        if attachments is not None:
            self.download_Attachments(objectId, attachments, Object)
        return attachments

    def download_Attachments(self, objectId, Attachments, Object=None):
        self.log.debug("  downloading attachment for '%s'" % objectId)
        if Object is not None:
            object_path = Object.find('path').text
        else:
            object_path = self.session.Ctx['project'].xml.find('path').text

        for attach in Attachments:
            xid = attach.find('attachmentId').text
            fileName = attach.find('fileName').text
            fileSize = int(attach.find('fileSize').text)
            if objectId.startswith('PageApp'):
                url_base_path = os.path.join('/sf/page/do/viewImage', object_path, 
                                             'page', fileName)
            if objectId.startswith('wiki'):
                url_base_path = os.path.join('/sf/wiki/do/viewAttachment', object_path, 
                                             fileName)

            file_data = self.HTTPDownloader.getFile(url_base_path)
            if len(file_data) != fileSize:
                self.log.warn("fileSize did not match recorded size for '%s'" % xid)
                self.log.warn(" expected size = %s, found = %s" % (fileSize, len(file_data)))
            file_path = os.path.join(self.base_dir, 'Attachments', objectId, fileName)
            self.add_to_archive(file_path, StringIO.StringIO(file_data))

    def export_PageComponents(self, page):
        pageId = page.find('id').text
        pc_list = ApiMethods['getPageComponents'](self.session,
                                                  'getPageComponents',
                                                  [pageId]).export_xml()
        if self.session.error > 0:
            self.log.error("failed to export PageComponent for page '%s'" % pageId)
            sys.exit(1)
        if pc_list is None:
            return

        for pc in pc_list:
            pcId = pc.find('id').text
            self.log.debug1("    processing pc '%s'" % pcId)
            page.append(pc)
            pctype = pc.find('componentType').text
            if pctype == 'html' or 'metrics':
                if pctype == 'html':
                    cmd = 'getTextComponentContent'
                else:
                    cmd = 'getPageComponent'
                content = ApiMethods[cmd](self.session, cmd, [pcId]).export_xml()
                if self.session.error > 0:
                    self.log.error("failed to export html pageComponent '%s'" % pcId)
                else:
                    self.log.debug("    exporting html pageComponent '%s'" % pcId)
                path = os.path.join(self.base_dir, 'Pages', pageId, pcId)
                if content is not None:
                    # content is html, create file-like object from it
                    buf = content.text
                else:
                    buf = ''
                    self.log.debug1("    component is empty, '%s'" % pcId)
                f = StringIO.StringIO(buf)
                self.add_to_archive(path, f)

    def export_TrackerApp(self, projectId):
        self.log.info("Exporting project trackers.")
        trackerApp = etree.Element('TrackerApp')
        trackers = etree.SubElement(trackerApp, 'Trackers')
        trackerList = []
        if projectId:
            cmd = 'getTrackerList'
            trackerList = ApiMethods[cmd](self.session, cmd, [projectId]).export_xml()
            if trackerList is not None :
                for t in trackerList:
                    trackerId = t.find('id').text
                    title = t.find('title').text

                    # get tracker data
                    cmd = 'getTrackerData'
                    tracker_data = ApiMethods[cmd](self.session, cmd, [trackerId]).export_xml()
                    if self.session.error > 0:
                        self.log.error("failed to export tracker definition for '%s'" % title)
                        sys.exit(1)
                    else:
                        self.log.verbose("  exporting tracker definition for '%s'" % title)

                    # get fields data
                    cmd = 'getFields'
                    tracker_fields = ApiMethods[cmd](self.session, cmd, [trackerId]).export_xml()
                    if self.session.error > 0:
                        self.log.error("failed to export tracker fields for '%s'" % title)
                        sys.exit(1)
                    else:
                        self.log.debug("    exporting tracker fields for '%s'" % title)
                    tracker_data.append(tracker_fields)

                    # get artifacts
                    tracker_data.append(self.export_Artifacts(trackerId))                   
                    trackers.append(tracker_data)
        else:
            self.log.error("must provide a projectId as argument.")
        self.log.info("Exporting all trackers completed")
        self.log.info("")
        return trackerApp

    def export_Artifacts(self, trackerId):
        artifacts = etree.Element('Artifacts')
        cmd = 'getArtifactList'
        argvs = [trackerId, '']
        artifact_list = ApiMethods[cmd](self.session, cmd, argvs).export_xml()
        if self.session.error > 0:
            self.log.error("failed to get artifact list for tracker '%s'" % trackerId)
            sys.exit(1)
        if len(artifact_list) > 0:
            cmd = 'getArtifactData'
            for artifact in artifact_list:
                artifactId = artifact.find('id').text
                title = artifact.find('title').text
                artifact_data = ApiMethods[cmd](self.session, cmd, [artifactId]).export_xml()
                if self.session.error > 0:
                    self.log.error("failed to get artifact data for '%s'" % artifactId)
                else:
                    self.log.debug("    exporting from tracker '%s' artifact '%s'" % \
                                   (trackerId, title))
                # get child dependencies
                artifact_data.append(self.export_ChildDependency(artifactId, title))

                artifacts.append(artifact_data)
        return artifacts

    def export_ChildDependency(self, artifactId, title):
        cmd = 'getChildDependencyList'
        children = ApiMethods[cmd](self.session, cmd, [artifactId]).export_xml()
        if self.session.error > 0:
            self.log.error("failed to get child dependency list for artifact '%s'" % artifactId)
        else:
            self.log.debug("    exporting child dependency list for artifact '%s'" % title)
        return children

    def export_PlanningApp(self, projectId):
        self.log.info("Exporting project planning folders.")
        if projectId:
            cmd1 = 'getPlanningFolderList'
            cmd2 = 'getPlanningFolderData'
            planningApp = etree.Element('PlanningApp')
            # get PlanningStatusValues if defined
            planningApp.append(self.export_PlanningStatusValues(projectId))
            pfs = etree.SubElement(planningApp, 'PlanningFolders')
            pfList = []
            pfList = ApiMethods[cmd1](self.session, cmd1, [projectId, 'true']).export_xml()
            if self.session.error > 0:
                self.log.error(" can't get planning folder list, or feature is not supported")
                self.log.info("")
                return planningApp
            if pfList is not None :
                for p in pfList:
                    pfId = p.find('id').text
                    self.log.verbose("  exporting planning folder '%s'." % pfId)
                    pf = ApiMethods[cmd2](self.session, cmd2, [pfId]).export_xml()
                    if self.session.error > 0:
                        self.log.error("Failed to export planning folder '%s'" % pfId)
                        sys.exit(1)
                    pfs.append(pf)
            self.log.info("Exporting all planning folders completed.")
        else:
            self.log.error("must provide a projectId as argument.")
        self.log.info("")
        return planningApp

    def export_PlanningStatusValues(self, projectId):
        self.log.debug("Exporting planningApp status values.")
        cmd = 'getPlanningStatusValues'
        status_val = ApiMethods[cmd](self.session, cmd, [projectId]).export_xml()
        if self.session.error > 0:
            self.log.error("Can't export planningApp status values '%s'" % projectId)
            sys.exit(1)
        return status_val

    def export_DocumentApp(self, projectId):
        self.log.info("Exporting project document folders.")
        docApp = etree.Element('DocumentApp')
        dfs = etree.SubElement(docApp, 'DocumentFolders')
        dfList = []
        if projectId:
            cmd = 'getDocumentFolderList'
            dfList = ApiMethods[cmd](self.session, cmd,
                                     [projectId, True]).export_xml()
            if self.session.error > 0:
                self.log.error("Can't retriev docfolder list for project '%s'" % projectId)
                sys.exit(1)
            for doc_folder in dfList:
                dfId = doc_folder.find('id').text
                self.log.verbose("  exporting document folder '%s'" % dfId)
                cmd = 'getDocumentFolderData'
                df_data = ApiMethods[cmd](self.session, cmd, [dfId]).export_xml()
                if self.session.error > 0:
                    self.log.error("Can't export doc folder '%s'" % dfId)
                    sys.exit(1)
                # investigate: file download efficiency
                # df_data.append(self.export_Documents(dfId))
                dfs.append(df_data)
        else:
            self.log.error("must provide a projectId as argument.")

        self.log.info("Exporting project document folders completed")
        self.log.info("")
        return docApp

    def export_Documents(self, dfId):
        #dfId = docfolder.find('id').text
        self.log.debug("Exporting documents in folder '%s'" % dfId)
        docs = etree.Element('Documents')
        cmd = 'getDocumentList'
        # alway provide an empty filter, to get complete doc list
        doclist = ApiMethods[cmd](self.session, cmd, [dfId, '']).export_xml()
        if self.session.error > 0:
            self.log.error("Can't get documnet list from folder '%s'" % dfId)
            sys.exit(1)

        for doc in doclist:
            docId = doc.find('id').text
            title = doc.find('title').text
            currentVersion = doc.find('currentVersion').text
            cmd = 'getDocumentData'
            argvs = [docId, currentVersion]
            doc_data = ApiMethods[cmd](self.session, cmd, argvs).export_xml()
            if self.session.error > 0:
                self.log.error("Can't export document '%s' in folder '%s'" % \
                                (docId, dfId))
                sys.exit(1)
            if doc_data is not None:
                self.log.debug("  exporting latest version of document '%s'" % title)
                docs.append(doc_data)
            # download file 
            self.download_DocContent(dfId, doc)
        return docs

    def download_DocContent(self, docFolderId, doc):
        docId = doc.find('id').text
        fileName = doc.find('fileName').text
        if fileName is None or fileName == '':
            return
        fileSize = int(doc.find('fileSize').text)
        currentVersion = doc.find('currentVersion').text
        cmd = 'getDocumentFileId'
        argvs = [docId, currentVersion]
        guid = ApiMethods[cmd](self.session, cmd, argvs).export_xml().text
        if self.session.error > 0:
            self.log.error("can't get GUID for document '%s'" % docId)
        cmd = 'read'
        argvs = [guid, 0, fileSize]
        content = ApiMethods[cmd](self.session, cmd, argvs).export_data()
        if len(content) != fileSize:
            self.log.error("file size differs from expected (%s), got %s" % \
                           (fileSize, len(content)))
        file_path = os.path.join(self.base_dir, 'Documents', docFolderId, docId, 
                                 '%s.%s' % (fileName, currentVersion))
        self.add_to_archive(file_path, StringIO.StringIO(content))
        self.log.debug("  downloading document: folderId=%s, filename=%s" % \
                         (docFolderId, fileName))
        
    def export_WikiApp(self, projectId):
        self.log.info("Exporting project Wiki pages.")
        wikiApp = etree.Element('WikiApp')
        wikipages = etree.SubElement(wikiApp, 'WikiPages')
        wp_list = []
        if projectId:
            cmd = 'getWikiPageList'
            argvs = [projectId]
            wp_list = ApiMethods[cmd](self.session, cmd, argvs).export_xml()
            if self.session.error > 0:
                self.log.error("Can't export wikipage list for project '%s'" % projectId)
                sys.exit(1)
            
            if wp_list is not None :
                for w in wp_list:
                    wpId = w.find('id').text
                    self.log.verbose("  exporting wikipage '%s'" % wpId)
                    cmd = 'getWikiPageData'
                    wp_data = ApiMethods[cmd](self.session, cmd, [wpId]).export_xml()
                    if self.session.error > 0:
                        self.log.error("Can't export wikipage data for '%s'" % wpId)
                        #sys.exit(1)
                        continue
                    # get attachments for each wikipage
                    wp_data.append(self.export_Attachments(wpId, wp_data))
                    wikipages.append(wp_data)
            else:
                self.log.warn("project %s does not have any wikipages" % projectId)

        self.log.info("Exporting wiki pages completed.")
        self.log.info("")
        return wikiApp

    def export_RbacApp(self, projectId):
        self.log.info("Exporting project roles.")
        rbacApp = etree.Element('RbacApp')

        cmd = 'getRoleList'
        role_list = ApiMethods[cmd](self.session, cmd, [projectId]).export_xml()
        if self.session.error > 0:
            self.log.error("failed to get role list for project '%s'" % projectId)

        if role_list is not None:
            cmd = 'listClusters'
            for role in role_list:
                roleId = role.find('id').text
                title = role.find('title').text
                clusters = ApiMethods[cmd](self.session, cmd, [roleId]).export_xml()
                if self.session.error > 0:
                    self.log.error("failed to get role list for project '%s'" % projectId)
                else:
                    self.log.verbose("  exporting roleCluster for role '%s'" % title)
                role.append(clusters)

        rbacApp.append(role_list)
        self.log.info("Exporting project roles completed.")
        self.log.info("")

        return rbacApp
               
    def export_FrsApp(self, projectId):
        self.log.info("Exporting project packages.")
        frsApp = etree.Element('FrsApp')
        pkgs = etree.SubElement(frsApp, 'Packages')

        cmd = 'getPackageList'
        pkg_list = ApiMethods[cmd](self.session, cmd, [projectId]).export_xml()
        if self.session.error > 0:
            self.log.error("failed to get package list for project '%s'" % projectId)

        if pkg_list is not None:
            cmd = 'getPackageData'
            for pkg in pkg_list:
                pkgId = pkg.find('id').text
                title = pkg.find('title').text
                pkg_data = ApiMethods[cmd](self.session, cmd, [pkgId]).export_xml()
                if self.session.error > 0:
                    self.log.error("failed to get pkg list for project '%s'" % projectId)
                else:
                    self.log.verbose("  exporting package '%s'" % title)
                pkg_data.append(self.export_Releases(pkgId, title))
                pkgs.append(pkg_data)

        self.log.info("Exporting project packages completed.")
        self.log.info("")

        return frsApp

    def export_Releases(self, pkgId, pkgTitle):
        releases = etree.Element('Releases')
        cmd = 'getReleaseList'
        release_list = ApiMethods[cmd](self.session, cmd, [pkgId]).export_xml()
        if self.session.error > 0:
            self.log.error("failed to get release list for package '%s'" % pkgId)
            sys.exit(1)
        if release_list is not None:
            cmd = 'getReleaseData'
            for release in release_list:
                releaseId = release.find('id').text
                title = release.find('title').text
                release_data = ApiMethods[cmd](self.session, cmd, [releaseId]).export_xml()
                if self.session.error > 0:
                    self.log.error("failed to get release list for project '%s'" % pkgId)
                else:
                    self.log.debug("    exporting release %s for package %s" % (title, pkgTitle))
                releases.append(release_data)
        return releases



