# -*- coding: utf-8 -*-

from lxml import etree
from ImportModule import ApiMethods
from SimpleLog import SimpleLog
import os
import re
import shutil
import sys
import tarfile
import tempfile
import DataUtils
import CTFApp

import time

# __revision__  = $LastChangedRevision$
# __Copyright__ = Copyright (c) 2007 - 2009, CollabNet, Inc. All rights reserved.

commands = ['load']
HomePageTitle = ['Project Home', '项目主页', 'プロジェクトホーム', '프로젝트 홈']

class load(object):
    def __init__(self, session, args=None):
        self.session = session
        self.loglevel = self.session.loglevel
        self.log = SimpleLog(self.loglevel)
        self.args = args
        self.base_dir = tempfile.mkdtemp('', 'tmp', '/tmp')
        self.projectId = None
        self.trackerId = None
        self.format = self.session.format
        self.data = None
        self.data_version = None
        self.data_apiversion = None
        self.min_version = '5.3.0.0'  # this is static data, don't change it when branching.
        self.old_new = {}
        self.new_old = {}
        self.map = {}
        self.nmap = {}
        self.nodes = {}
        self.nnodes = {}
        self.page_count = 0
        self.page_total = 0
        self.subfolder_count = 0
        self.subfolder_total = 0
        self.pf_total = 0
        self.pf_count = 0
        self.df_total = 0
        self.df_count = 0
        self.path_map = {}      # key, value => (old folder path, new folder path)
        self.orig_projectName = None
        self.known_projects = {}
        self.known_templates = {}
        self.known_project_urlnames = {}
        self.known_template_urlnames = {}
        
    def usage(self, errmsg=None):
        if errmsg:
            self.log.error("%s" % errmsg)
        self.log.info("Recreating project from data read from a XML file:")
        self.log.info("Usage: load inputfile")

    def cleanup(self, tmpdir):
        try:
            shutil.rmtree(tmpdir)
        except IOError, e:
            self.log.error("Failed to cleanup tmpdir '%s'" % tmpdir)
            self.log.error("Error message: %s" % e)

    def validate(self):
        if not self.session.session_id:
            self.log.warn("No active session, please login to server first")
            return False

        if self.args and len(self.args) != 1:
            self.usage("syntax error.")
            return False

        if not os.path.exists(self.args[0]):
            self.log.error("file '%s' does not exist." % self.args[0])
            return False

        if not tarfile.is_tarfile(self.args[0]):
            self.log.error("file format error, file '%s' is not a gzipped tar archive.")
            return False

        try:
            tar = tarfile.open(self.args[0], 'r:gz')
            members = tar.getmembers()
            for m in members:
                tar.extract(m, self.base_dir)
            tar.close()
        except IOError, ie:
            self.log.error("can't extract content from '%s'." % self.args[0])
            self.log.error("error message : %s" % repr(ie))
            self.logoff_exit(1)

        project_data = os.path.join(self.base_dir, 'project.xml')
        self.data = etree.parse(project_data)
        root = self.data.getroot()
        self.data_version = root.get("Version")
        self.data_apiversion = root.get("ApiVersion")
        if not DataUtils.isSameOrNewer(self.data_apiversion, self.session.ApiVersion):
            self.log.error("Cannot load data into older version TeamForge server.")
            self.log.error("Minimum TeamForge version required is '%s'." % self.data_version)
            return False

        return True
        
    def run(self):
        if not self.validate():
            return
        self.load_ProjectData(self.data)
        self.set_AccessLevel(self.data)
        self.load_DocumentApp(self.data.find('.//DocumentFolders'))
        self.load_WikiApp(self.data.find('.//WikiApp'))
        self.load_RbacApp(self.data.find('.//Roles'))
        # FrsApp comes before PlanningApp, which in turn comes before TrackerApp
        self.load_FrsApp(self.data.find('.//Packages'))
        self.load_PlanningApp(self.data.find('.//PlanningApp'))
        self.load_TrackerData(self.data.find('.//Trackers'))
        self.create_Dependencies()
        # 
        # NOTE: pageApp must be added after other Apps, since it need access to 
        # other components, like doc folders, wiki pages, news postings etc.
        #
        self.load_PageApp(self.data.find('.//PageApp'))
        self.create_Template()
        self.log.info("All project data loaded successfully.")
        self.cleanup(self.base_dir)
        # Close of stderr file descriptor causes program to terminate
        #os.close(sys.stderr.fileno())

    def load_ProjectData(self, data):
        """Create new project with same name/title/description. If duplicates
        are found, new project with have a number attached to name and title"""
        self.log.info("Creating project definition ...")
        project = data.find('Project')
        name = project.find('path').text.split('.')[1]
        title = project.find('title').text
        description = project.find('description').text
        self.orig_projectName = name

        # check to see if a project/template with identical name/title already exists
        self.get_projects_and_templates_info()
        incr = 0
        new_name = name
        new_title = title
        while new_title in self.known_projects or \
              new_title in self.known_templates or \
              new_name in self.known_project_urlnames or \
              new_name in self.known_template_urlnames:
            incr += 1
            new_name = name + '_' + str(incr)
            new_title = title + " %s" % str(incr)
  
        if title != new_title :
            self.log.warn("Another project or template already exists with the same title")
            self.log.warn("Renaming this project to '%s'" % new_title)

        new_project = ApiMethods['createProject'](self.session, 
                                 'createProject',
                                 [new_name, new_title, description]).export_xml()

        if new_project is not None:
            self.session.Ctx['project'] = CTFApp.Project(new_project)
            self.projectId = self.session.Ctx['project'].id
            self.log.info("Finished creating project definition successfully.")
            self.log.info("")
        else:
            self.log.error("Failed creating project definition, aborting.")
            self.logoff_exit(1)

    def get_projects_and_templates_info(self):
        """List all projects and templates, save their title and url names in maps"""
        cmd = 'getProjectList'
        projects = ApiMethods[cmd](self.session, cmd, []).export_xml()
        if self.session.error > 0:
            self.log.error("failed to list all projects")
        for p in projects.iter('Project'):
            title = p.find('title').text
            self.known_projects[title] = 1 
            p_path = p.find('path').text
            p_urlname = p_path.split('.')[1]
            self.known_project_urlnames[p_urlname] = 1
            
        cmd = 'listTemplates'
        templates = ApiMethods[cmd](self.session, cmd, []).export_xml()
        if self.session.error > 0:
            self.log.error("failed to list templates with title '%s'" % title)
        for t in templates.iter('Template'):
            title = t.find('title').text
            self.known_templates[title] = 1 
            t_path = t.find('path').text
            t_urlname = t_path.split('.')[1]
            self.known_template_urlnames[t_urlname] = 1
        
    def set_AccessLevel(self, data):
        self.log.info("Setting project access level.")
        if self.session.Ctx['project'] is None:
            self.log.error("Project context is not set, aborting.")
            self.logoff_exit(1)

        access_level = data.find('AccessLevel').text
        ApiMethods['setProjectAccessLevel'](self.session, 
                                            'setProjectAccessLevel',
                                            [access_level]).run()
        if self.session.error > 0:
            self.log.error("Failed setting project access level.")
        else:
            self.log.info("Finished setting project access level successfully.")
        self.log.info("")

    def load_TrackerData(self, data):
        if data is None:
            self.log.info("There are no trackers to be created for this project.")
            self.log.info("")
            return True
        if self.session.Ctx['project'] is None:
            self.log.error("Project context is not set. Aborting load_TrackerData().")
            self.logoff_exit(1)
        self.log.info("Creating project trackers ...")
        n = 0 
        for tracker in data.iter('Tracker'):
            path = tracker.find('path').text
            title = tracker.find('title').text
            old_trackerId = tracker.find('id').text
            description = tracker.find('description').text
            old_fields = tracker.find('Fields')
            artifacts = tracker.find('Artifacts')

            m = re.match(r'projects.(.*)/tracker.(.*)', path)
            if m and m.group(2) is not None:
                name = m.group(2)
            if not name:
                self.log.error("failed to determine tracker name")
                self.logoff_exit(1)
            cmd = 'createTracker'
            icon = tracker.find('icon').text
            argvs = [name, title, description, icon]
            new_tracker = ApiMethods[cmd](self.session, cmd, argvs).export_xml()
            if self.session.error > 0:
                self.log.error("Failed to create tracker '%s'" % title)
                self.logoff_exit(1)
            else:
                n += 1

            if new_tracker is not None:
                self.session.Ctx['tracker'] = CTFApp.Tracker(new_tracker)
                self.trackerId = new_tracker.find('id').text
                self.old_new[old_trackerId] = self.trackerId
                self.log.verbose("  created tracker '%s'." % name)
                cmd = 'getFields'
                time.sleep(1)
                new_fields = ApiMethods[cmd](self.session, cmd, [self.trackerId]).export_xml()
            else:
                self.log.error("Failed to create tracker '%s'." % name)
                self.logoff_exit(1)

            self.update_Fields(old_fields, new_fields)
            self.add_Artifacts(artifacts, self.trackerId)

        self.log.info("Finished creating all trackers, total = %s." % n)
        self.log.info("")

    def update_Fields(self, old, new):
        """Update new tracker's field settings based on old tracker
           configuration
        Note: old is in xml format, while new is an array of TrackerFieldSoapDO object!
        """
        of_dict = {}
        nf_dict = {}

        if 'tracker' not in self.session.Ctx or self.session.Ctx['tracker'] is None:
            self.log.error("Tracker context is not set. Aborting set_Fields().")
            self.logoff_exit(1)
        if old is None or new is None:
            self.log.info("There are no fields to be created for this tracker.")
            return True

        # create maps {field_name: field_obj} for both old and new tracker fields
        for o_field in old.iter('Field'):
            of_name = o_field.find('name').text
            of_dict[of_name] = o_field
        for n_field in new.iter('Field'):
            nf_name = n_field.find('name').text
            nf_dict[nf_name] = n_field
            xid = n_field.find('id').text
            self.session.trackerFields[xid] = n_field
        for key in of_dict.keys():
            of = of_dict[key]
            if key in nf_dict:
                nf = nf_dict[key]
                self.copy_set_field(of, nf)
            else:
                # this is not a default field, so not present in new tracker
                self.add_field(of)

    def copy_set_field(self, field1, field2):
        """Selectively copy some attribute values from field1 to the field2"""
        field_data = {}
        excluded = ['createDate', 'createdBy', 'id', 'lastModfiedBy', 
                    'lastModifiedDate', 'version']

        nf_id = field2.find('id').text
        fieldName = field2.find('name').text
        self.log.debug5("      =========================================")
        self.log.debug5("      updating field '%s'" % fieldName)

        # copy values from old field settings
        for child in field1:
            tag = child.tag
            if tag in excluded or tag == 'Field':
                continue
            if tag == 'fieldValues':
                values_tup = self.get_fieldValues(child)
                field_data[tag] = values_tup
            else:
                field_data[tag] = child.text

        # copy values from new field settings
        for child in field2:
            tag = child.tag
            if tag in excluded:
                if tag == 'version':
                    # new fields always have version = 0
                    field_data[tag] = 0
                else:
                    field_data[tag] = child.text

        field_data['id'] = nf_id
        if DataUtils.isSameOrNewer(self.min_version, self.data_version):
            cmd = 'setField'
        else:
            cmd = 'setField'
        argvs = [self.trackerId, str(field_data)]

        ApiMethods[cmd](self.session, cmd, argvs).run()
        if self.session.error > 0:
            self.log.error("Can't set field data for tracker '%s', field '%s'" % \
                           (self.trackerId, fieldName))
        else:
            self.log.debug("    updated field data for tracker '%s', field '%s'" % \
                           (self.trackerId, fieldName))

    def add_field(self, field):
        trackerTitle = self.session.Ctx['tracker'].title
        columns = field.find('displaySize').text
        #defaultTextValue = field.find('defaultTextValue').text
        hidden = DataUtils.to_boolean(field.find('hiddenOnCreate').text)
        lines = field.find('displayLines').text
        name = field.find('name').text
        disabled = DataUtils.to_boolean(field.find('disabled').text)
        required = DataUtils.to_boolean(field.find('required').text)
        fieldtype = field.find('fieldType').text
        self.log.debug5("      =========================================")
        self.log.debug5("      adding new '%s' type field '%s'" % (fieldtype, name))
        values_tup = self.get_fieldValues(field.find('fieldValues'))
        values = []
        defaultval = []
        for tup in values_tup:
            values.append(tup[0])
            if tup[2]:
                defaultval.append(tup[0])
        # Notes:
        # if a field isRequired = True, then it must have a defaul value
        # which, in turn, means that values list must not be empty or None
        if fieldtype == "single-select":
            cmd = 'addSingleSelectField'
            if len(defaultval) > 0:
                d = defaultval[0]
            else:
                d = 'none'
            argvs = [name, required, disabled, hidden, values, d]
        elif fieldtype == 'date':
            cmd = 'addDateField'
            argvs = [name, required, disabled, hidden]
        elif fieldtype == 'multi-select':
            cmd = 'addMultiSelectField'
            argvs = [name, lines, required, disabled, hidden, values, defaultval]
        elif fieldtype == 'multi-select-user':
            cmd = 'addUserSelectField'
            # if require is True, then defaultNames can't be empty!
            defaultNames = field.find('defaultUsernames').text
            userfilter = field.find('userFilter').text
            argvs = [name, lines, required, disabled, hidden, defaultNames, userfilter]
        elif fieldtype == 'text':
            cmd = 'addTextField'
            # if required == True, there must be a default value!
            if len(defaultval) > 0:
                d = defaultval[0]
            else:
                if required:
                    d = 'put default value here'
                else:
                    d = ''
            argvs = [name, columns, lines, required, disabled, hidden, d]

        ApiMethods[cmd](self.session, cmd, argvs).export_xml()
        if self.session.error > 0:
            self.log.error("Failed to create tracker field '%s' for tracker '%s'" \
                      % (name, trackerTitle))
            self.log.error("(add_field): field type '%s', args: %s" % (fieldtype, argvs))
            if cmd != 'addUserSelectField':
                self.logoff_exit(1)
        else:
            self.log.debug("      added new '%s' type field '%s'" % (fieldtype, name))

    def get_fieldValues(self, fieldValues):
        """return field values as a list of tuples"""
	# note: 
	#  fieldValue has three attributes: value, valueClass, isDefault
	#
	#  when adding new fields, valueClass is not specified. However, when setting fieldData
	#  all three attributes must be set properly
	#     
        val = []
        if fieldValues is None:
            return None
        for f in fieldValues.iter('fieldValue'):
            v = f.find('value').text
            isDefault = f.find('isDefault').text
            valClass = f.find('valueClass').text
            if isDefault is None:
                isDefault = False
            val.append((v, valClass, isDefault))
        self.log.debug5("    field values: %s" % val)
        return val

    def add_Artifacts(self, artifacts, trackerId):
        if artifacts is None or len(artifacts) == 0:
            self.log.verbose("    there are no artifacts to load for tracker '%s'" % trackerId)
            self.log.verbose("")
            return
        self.log.verbose("    adding artifacts")
        n = 0
        cmd = 'createArtifact'
        for artifact in artifacts:
            actualEffort = 0
            actualHours = 0
            aid = artifact.find('id').text
            title = artifact.find('title').text
            description = artifact.find('description').text
            group = artifact.find('group').text
            category = artifact.find('category').text
            status = artifact.find('status').text
            customer = artifact.find('customer').text
            priority = int(artifact.find('priority').text)
            # FIXME: once we move over user accounts, fix this
            #assignedTo = artifact.find('assignedTo').text
            assignedTo = 'nobody'
            #releaseId = artifact.find('reportedReleaseId').text
            releaseId = 'none'
            flexFields = str(self.get_flexFields(artifact.find('flexFields')))

            # attachment to be added later
            #attachmentFileName = artifact.find('attachmentFileName').text
            #attachmentMimeType = artifact.find('attachmentMimeType').text
            #attachmentFileId = artifact.find('attachmentFileId').text
            a = artifact.find('actualEffort').text
            if a is None:
                actualEffort = 0
            else:
                actualEffort = int(a)
            autosumming = DataUtils.to_boolean(artifact.find('autosumming').text)
            if autosumming:
                estimatedEffort = 0
                remainingEffort = 0
            else:
                e = artifact.find('estimatedEffort').text
                r = artifact.find('remainingEffort').text
                if e is None:
                    estimatedEffort = 0
                else:
                    estimatedEffort = int(e)
                if r is None:
                    remainingEffort = 0
                    estimatedEffort = 0
                else:
                    remainingEffort = int(r)
            old_planningFolderId = artifact.find('planningFolderId').text
            if old_planningFolderId is not None and old_planningFolderId in self.old_new:
                planningFolderId = self.old_new[old_planningFolderId]
            elif old_planningFolderId is None:
                planningFolderId = None
            else:
                planningFolderId = None
                self.log.error("matching planning folder not found in new project")
                self.log.error("original planning folder title: '%s'" % title)

            p = artifact.find('points')
            if p is not None:
                points = int(p.text)
            else: 
                points = 0
            argvs = [trackerId, title, description, group, category, status, customer,
                     priority, estimatedEffort, remainingEffort, autosumming, points,
                     assignedTo, releaseId, planningFolderId, flexFields,
                     '', '', '']
            new_artifact = ApiMethods[cmd](self.session, cmd, argvs).export_xml()
            if self.session.error > 0:
                self.log.error("    can't create artifact '%s'" %  title)
                #sys.exit(1)
            else:
                n += 1
                new_aid = new_artifact.find('id').text
                self.session.artifacts[new_aid] = new_artifact
                self.old_new[aid] = new_aid
                self.log.debug("    created artifact '%s'" % title)

                if not autosumming and (actualEffort or actualHours):
                    cmd = 'setArtifactData'
                    argvs = ["{'id':'%s', 'actualEffort':%d}" % (new_aid, actualEffort),
                                 '', '', '', '']
                    ApiMethods[cmd1](self.session, cmd, argvs).run()
                    if self.session.error > 0:
                        self.log.error("    can't set artifactData for '%s'" %  title)
                        #sys.exit(1)
                    else:
                        self.log.debug("    updated actualEffort for artifact '%s'" %  title)

        self.log.verbose("    finished adding artifacts, total = %s." % n)

    def create_Dependencies(self):
        """Create artifacts dependencies."""
        self.log.info("Creating artifacts dependencies")
        cmd = 'createArtifactDependency'
        
        for artifact in self.data.iter('Artifact'):
            if len(artifact) == 0:
                continue
            aid = artifact.find('id').text
            children = artifact.find('ChildDependencys')
            if len(children) == 0:
                continue
            if aid not in self.old_new:
                self.log.error("artifacts '%s' not found in new project" % aid)
                continue
            for child in children.iter('ChildDependency'):
                targetId = child.find('targetId').text
                if targetId not in self.old_new:
                    self.log.error("artifacts '%s' not found in new project" % targetId)
                    continue
                desc = child.find('description').text
                argvs = [self.old_new[aid], self.old_new[targetId], desc]
                ApiMethods[cmd](self.session, cmd, argvs).run()
                if self.session.error > 0:
                    self.log.error("failed creating dependency for '%s': '%s'" % \
                                   (aid, targetId))
                else:
                    self.log.debug("    creating dependency for '%s': '%s'" % \
                                  (aid, targetId))
        self.log.info("Finished creating artifacts dependencies")
        self.log.info("")
              
    def get_flexFields(self, flexFields):
        """Returns a list of three lists: [names, types, values]"""
        ff = []
        names = flexFields.find('names').text
        types = flexFields.find('types').text
        values = flexFields.find('values').text
        if names:
            names = eval(names)
            types = eval(types)
            values = eval(values)
            if types:
                for i in range(len(types)):
                    if types[i] == 'User':
                        values[i] = None
        ff.append(names)
        ff.append(types)
        ff.append(values)
        return ff

    def load_PlanningApp(self, data):
        if data is None:
            self.log.warn("There are no planning folders to load.")
            self.log.info("")
            return
        else:
            pfs = data.find("PlanningFolders")
            status_values = data.find("PlanningStatusValues")
            self.pf_total = len(pfs)

        self.log.info("Creating project planning folders ...")
        self.subfolder_count = 0

        # PlanningStatusValues only supported in CTF 5.4 or above
        if status_values is not None and \
           DataUtils.isSameOrNewer('5.4.0.0', self.session.Version):
            self.log.debug("  updating planning app status values ...")
            self.update_status_values(status_values)

        # The only way to extract PlanningAppXXXX is to get parent_path of a planningFolder,
        # if there are no planningFolders, we get None
        oldPfApp, self.map, self.nodes = DataUtils.find_children(pfs)
        if oldPfApp is None:
            self.log.warn("This project has no project planning folders to import")
            return 

        for pid in self.map[oldPfApp]:
            self.add_subfolders(pid, self.projectId, 'PlanningFolder')

        self.pf_count = self.subfolder_count
        if self.pf_total == self.pf_count: 
            self.log.info("Finished creating all project planning folders, total = %s." \
                  % self.pf_count)
        else:
            self.log.warn("Expecting %s planning folders, %s created." % \
                      (self.pf_total, self.pf_count))
        self.log.info("")

    def update_status_values(self, values):
        """values is in XML format, need to convert into list of tuples for
           setPlanningStatusValues call"""
        sv = []
        for item in values:
            value = item.find('value').text
            valueClass = item.find('valueClass').text
            isDefault = DataUtils.to_boolean(item.find('isDefault').text)
            sv.append((isDefault, value, valueClass))
        cmd = 'setPlanningStatusValues'
        args = [self.projectId, sv]
        ApiMethods[cmd](self.session, cmd, args).export_xml()
        if self.session.error > 0:
            self.log.error("  updating planning app status values failed for '%s'" % self.projectId)
            self.logoff_exit(1)
        else:
            self.log.debug("  updating planning app status values for '%s'" % self.projectId)

    def load_DocumentApp(self, data):
        if data is None:
            self.log.warn("There are no document folders to load.")
            self.log.info("")
            return
        else:
            self.df_total = len(data)
        self.log.info("Creating project document folders ...")
        self.subfolder_count = 0

        oldDfAppId, self.map, self.nodes = DataUtils.find_children(data)
        oldRootId = self.map[oldDfAppId][0]
        old_path = self.nodes[oldRootId].find('path').text
        cmd = 'getDocumentFolderList'
        argvs = [self.projectId, False]
        newDocRoot = ApiMethods[cmd](self.session, cmd, argvs).export_xml()
        # there is only one docFolder for new project, it's the 'Root Folder'
        newRootId = newDocRoot.find('.//id').text
        new_path = newDocRoot.find('.//path').text
        self.path_map[old_path] = new_path
        self.log.debug2("    old docFolder path = '%s'" % old_path)
        self.log.debug2("    new docFolder path = '%s'" % new_path)

        self.old_new[oldRootId] = newRootId
        if oldRootId not in self.map:
            self.log.warn("This project has no project document folders to create.")
            self.log.info("")
            return 
            
        for dfId in self.map[oldRootId]:
            self.add_subfolders(dfId, newRootId, 'DocumentFolder')

        self.df_count = self.subfolder_count
        if self.df_total == self.df_count + 1: 
            self.log.info("Finished creating all project document folders, total = %s." \
                  % self.df_count)
        else:
            self.log.warn("Some document folders could not be created.")
            self.log.warn("Expecting %s doc folders, %s created." % \
                      (self.df_total, self.df_count))
        self.log.info("")

    def load_WikiApp(self, wikiapp):
        n = 0
        if wikiapp is None:
            self.log.warn("There are no wiki pages in exported project data.")
            self.log.info("")
            return 
        else:
            wikipages = wikiapp.find('WikiPages')
            num_wiki = len(wikipages)

        if wikipages is None:
            self.log.warn("This project has no wiki pages to import")
            return 
        self.log.info("Creating wiki pages ...")
        # all new project comes with two wiki pages: a "Project Home Page" and 
        # a "Home page". Therefore, they have to be treated separately.
        cmd = 'getWikiPageList'
        existing_pages = ApiMethods[cmd](self.session, cmd, [self.projectId]).export_xml()
        default_wikiPages = {}  # k, v = title, id
        for page in existing_pages:
            pId = page.find('id').text
            cmd = 'getWikiPageData'
            wpSoapData = ApiMethods[cmd](self.session, cmd, [pId]).export_data()
            title = wpSoapData.get_element_title()
            default_wikiPages[title] = pId
            self.session.wikiPages[title] = wpSoapData
            if title == 'Home page':
                default_wikiPages['HomePage'] = pId
                self.session.wikiPages['HomePage'] = wpSoapData
        for wp in wikipages:
            wikiId = wp.find('id').text
            title = wp.find('title').text
            attachments = wp.find('Attachments')

            if title in default_wikiPages:
                self.update_WikiPageData(wp, self.session.wikiPages[title])
                self.old_new[wikiId] = default_wikiPages[title]
                new_wp = wp
            else:
                versionComment = wp.find('versionComment').text
                wikiText = self.update_wikiText(wp.find('wikiText').text)
                # to create new wikiPage, wikiText can not be empty
                if wikiText is None or wikiText == '':
                    wikiText = "This is an empty page"
    
                cmd = 'createWikiPage'
                argvs = [self.projectId, title, wikiText, versionComment]
                new_wp = ApiMethods[cmd](self.session, cmd, argvs).export_xml()
                if self.session.error > 0:
                    self.log.error("  failed to create wikipage '%s'" % title)
                    self.logoff_exit(1)
                else:
                    n += 1
                    self.log.verbose("  created wiki page '%s'" % title)

            new_wpId = new_wp.find('id').text
            self.old_new[wikiId] = new_wpId
            
            if attachments is not None and len(attachments) >0 :
                new_wpId = new_wp.find('id').text
                self.add_Attachments(attachments, wikiId, new_wpId)

        if n == num_wiki - 2:
            self.log.info("Finished creating all wiki pages successfully, total = %s." % n)
        else:
            self.log.warn("Some wiki pages could not be recreated.")
            self.log.warn("Expecting: %s, created: %s." % (num_wiki-1, n))
        self.log.info("")

    def update_wikiText(self, text):
        if text is None or text == '':
            return ''
        found = []
        objId_list = re.findall('(\w+\d+)', text)
        if len(objId_list) == 0:
            return text
        for objId in objId_list:
            if objId in self.old_new:
                text = text.replace('%s' % objId, 'OLD:%s' % objId)
                found.append(objId)
        for objId in found:
            text = text.replace('OLD:%s' % objId, self.old_new[objId])
        return text

    def update_WikiPageData(self, oldPage, newPage):
        if oldPage is None or newPage is None:
            self.log.warn("    update_WikiPageData: one of two pages is not defined.")
            return
        # selectively update following fields:
        title = oldPage.find('title').text
        xid = newPage.get_element_id()
        versionComment = oldPage.find('versionComment').text
        wikiText = self.update_wikiText(oldPage.find('wikiText').text)
        if wikiText is None or wikiText == '':
            wikiText = "This is an empty page"
        cmd = 'setWikiPageData'
        argvs = ["{'id':'%s', 'title':'%s', 'versionComment':'%s', 'wikiText':'''%s'''}" % \
                   (xid, title, versionComment, wikiText)]
        ApiMethods[cmd](self.session, cmd, argvs).export_xml()
        if self.session.error > 0:
            self.log.error("  updating wiki page data failed for '%s'" %title)
        else:
            self.log.debug("  updating wiki page data for '%s'" %title)


    def load_PageApp(self, pageapp):
        if pageapp is None:
            self.log.warn("There are no project pages to load.")
            self.log.info("")
            return 
        data = pageapp.find('Pages')
        attachments = pageapp.find('Attachments')

        self.log.info("Creating project pages ...")
        self.page_total = len(data)
        oldPageAppId, self.map, self.nodes = DataUtils.find_children(data)

        # for newly created project, there is only one page, Project Home page
        ndata = ApiMethods['getPageList'](self.session, 'getPageList',
                                         [self.projectId]).export_xml()
        newPageAppId, self.nmap, self.nnodes = DataUtils.find_children(ndata)
        self.old_new[oldPageAppId] = newPageAppId
        self.new_old[newPageAppId] = oldPageAppId

        self.add_Attachments(attachments, oldPageAppId, newPageAppId)

        old_homepage_id = self.find_ProjectHome_pageId(data)
        new_homepage_id = self.find_ProjectHome_pageId(ndata)
        self.old_new[old_homepage_id] = new_homepage_id
        self.new_old[new_homepage_id] = old_homepage_id

        # need to remove all pagecomponents from new project's homepage
        if new_homepage_id:
            self.delete_PageComponents(new_homepage_id)

        for pid in self.map[oldPageAppId]:
            self.add_pages(pid, newPageAppId)

        if self.page_total == self.page_count + 1:
            self.log.info("Finished creating all project pages, total = %s." \
                  % self.page_count)
        else:
            self.log.warn("Expecting %s pages, %s pages created." % \
                      (self.page_total, self.page_count))
        self.log.info("")

    def find_ProjectHome_pageId(self, pages):
        for page in pages:
            if page.find('title').text in HomePageTitle:
                return page.find('id').text

    def add_Attachments(self, attachments, oldObjectId, newObjectId):
        if attachments is None or len(attachments) == 0:
            self.log.warn("    there are no attachment to add to '%s'" % newObjectId)
            return
        else:
            self.log.verbose("  created attachments to '%s'" % newObjectId)

        for atch in attachments:
            fileName = atch.find('fileName').text
            fileSize = int(atch.find('fileSize').text)
            mimetype = atch.find('mimetype').text
            file_path = os.path.join(self.base_dir, 'ProjectFiles/Attachments', 
                                     oldObjectId, fileName)
            try:
                f = open(file_path, 'r')
                content = f.read()
                f.close()
            except IOError, e:
                self.log.error("Can't read from file '%s'" % file_path)
                self.log.error("error message '%s'" % e)
                continue
            if len(content) != fileSize:
                self.log.error("file size differs from expected value for '%s'" % fileName)
                self.log.error("expected size = %s, found = '%s'" % (fileSize, len(content)))
                continue

            #start upload
            # get a file descriptor/guid
            cmd = 'startFileUpload'
            guid = ApiMethods[cmd](self.session, cmd, []).export_data()
            if guid is None:
                self.log.error("failed to obtain an in-session fileId for '%s'" % fileName)
                continue
            ApiMethods['write'](self.session, 'write', [guid, content]).run()
            if self.session.error > 0:
                self.log.error("failed to write to in-session fileId for '%s'" % fileName)
                continue
            ApiMethods['endFileUpload'](self.session, 'endFileUpload', [guid]).run()
            if self.session.error > 0:
                self.log.error("failed to cloes in-session fileId '%s'" % guid)
                continue

            # add attachment to new object:
            if newObjectId.startswith('PageApp'):
                cmd = 'addPageImage'
                argvs = [self.projectId, guid, fileName, mimetype]
            else:
                cmd = 'addAttachment'               # for wikiPages attachments
                argvs = [newObjectId, guid, fileName, mimetype]
            ApiMethods[cmd](self.session, cmd, argvs).export_xml()
            if self.session.error > 0:
                self.log.error("failed to add attachment to '%s'" % newObjectId)
                continue
            else:
                self.log.debug("    added attachment '%s' to '%s'" % (fileName, newObjectId))

    def delete_PageComponents(self, pageId):
        cmd = 'getPageComponents'
        pc_list = ApiMethods[cmd](self.session, cmd, [pageId]).export_xml()
        if self.session.error > 0:
            self.log.error("can't get pageComponent list for page '%s'" % pageId)
            self.logoff_exit(1)
        if pc_list is not None:
            cmd = 'deletePageComponent'
            for pc in pc_list:
                pcId = pc.find('id').text
                ApiMethods[cmd](self.session, cmd, [pcId]).run()
                if self.session.error > 0:
                    self.log.error("can't remove pageComponents '%s' from page '%s'" \
                                    % (pageId, pcId))
                    self.logoff_exit(1)

    def add_pages(self, oldPid, newParentId):
        """Add children pages to the new project PageApp"""
        n_homepageId = None

        # get pageId for new project's Porject Home page
        otitle = self.nodes[oldPid].find('title').text
        oParentId = self.nodes[oldPid].find('parentFolderId').text

        if oParentId not in self.old_new:
            self.old_new[oParentId] = newParentId
        if newParentId not in self.new_old:
            self.new_old[newParentId] = oParentId

        if otitle in HomePageTitle:
            for nid in self.nmap[newParentId]:
                npage = self.nnodes[nid]
                nparentId = npage.find('parentFolderId').text
                ntitle = npage.find('title').text
                if ntitle in HomePageTitle:
                    n_homepageId = nid
                self.old_new[oldPid] = n_homepageId
                self.new_old[n_homepageId] = oldPid

        page = self.nodes[oldPid]
        parentId = page.find('parentFolderId').text
        title = page.find('title').text
        visible = page.find('visible').text
        if not visible:
            visible = False
        if (page.find('freeform') is not None):
            freeform = page.find('freeform').text
        else:
            freeform = False

        if parentId not in self.old_new:
            self.log.error("parentId not in old_new association map, %s" % parentId)
            self.log.error("you should not reach here!")
            return

        if title not in HomePageTitle:
            new_page = ApiMethods['createPage'](self.session, 'createPage',
                       [self.projectId, self.old_new[parentId], title, visible, freeform]).export_xml()
            new_pageId = new_page.find('id').text
            nparentId = new_page.find('parentFolderId').text
            self.old_new[oldPid] = new_pageId
            self.new_old[new_pageId] = oldPid
            if nparentId not in self.nmap:
                self.nmap[nparentId] = []
            self.nmap[nparentId].append(new_pageId)
            self.nnodes[new_pageId] = new_page
            self.page_count += 1
            self.log.verbose("  created project page '%s'" % title)

        self.add_pageComponents(page)

        if oldPid in self.map and self.map[oldPid] is not None:
            for xid in self.map[oldPid]:
                if  oldPid in self.old_new:
                    self.add_pages(xid, self.old_new[oldPid])
                else:
                    self.log.error("parentId not found in old_new association map, '%s'" % oldPid)
                    self.log.error("You should not get this error :( !")
                        
    def add_pageComponents(self, page):
        """Add page components to a page"""

        pageId = page.find('id').text
        for pc in page.iter('PageComponent'):
            if pc is None:
                continue
            pctype = pc.find('componentType').text
            title = pc.find('title').text
            pcId = pc.find('id').text
            # Gosh: a typo in field name: visiblity !
            visibility = int(pc.find('visiblity').text)
            if (pc.find('topOfPage') is not None):
                topOfPage = pc.find('topOfPage').text
            else:
                topOfPage = False
            self.log.debug5("add_pageComponents: processing %s" % pcId)
            if pctype == 'html' or 'metrics':
                if not pageId in self.old_new:
                    self.log.error("Can't find this page in new project, '%s'" % pageId)
                    self.logoff_exit(1)
                fpath = os.path.join(self.base_dir, 'ProjectFiles/Pages', pageId, pcId)
                try:
                    f = open(fpath, 'r')
                    #content = f.read()
                    content = self.update_content(f.read())
                    f.close()
                except IOError, e:
                    self.log.warn("Can't read from file '%s'" % fpath)
                    self.log.warn("error message '%s'" % e)
                    content = ''
                attributes = ""
            elif pctype == 'graphs':
                attributes = ""
                content = ''
            elif pctype == 'documents':
                op = eval(pc.find('./settings/values').text)
                old_path = op[0]
                self.log.debug2("    old:  %s" % old_path)
                if old_path in self.path_map:
                    new_path = self.path_map[old_path]
                else:
                    self.log.warn("Document Folder for this component is not found in project data")
                    continue
                attributes = "['documentFolder'], ['%s']" % new_path
                content = ''
                self.log.debug2("    old, new:  %s => %s" % (old_path, new_path))
            elif pctype == 'wiki':
                values = eval(pc.find('./settings/values').text)
                value = values[0]
                attributes = "['wikiTopic'], ['%s']" % value
                content = ''
# FIXME: artf69916
#            elif pctype == 'chart':
#                names = eval(pc.find('./settings/names').text)
#                values = eval(pc.find('./settings/values').text)
#                #value = values[0]
#                attributes = "%s, %s" % (names, values)
#                content = ''
            else:
                continue
            cmd = 'createComponent'
            if pctype == 'metrics':
                cmd = 'createComponent2'
                attributes = ""
            # createComponent pageId compName componentType topOfPage visibility 
            #                 attributes content
            argvs = [self.old_new[pageId], title, pctype, topOfPage, visibility, attributes, content]
            ApiMethods[cmd](self.session, cmd, argvs).export_xml()
            if self.session.error > 0:
                self.log.error("can't create page component for page '%s'" % pageId)
                self.logoff_exit(1)
            else:
                self.log.debug("    created '%s' type pageComponent (old Id: '%s')" % (pctype, pcId))
           
    def update_content(self, content):
        """update intraproject url links on all page text content.
           fileObject can be any file-like objects supporting read/readline/realines
           methods"""
        found = []
        if content is None or len(content) == 0:
            return content
        objId_list = re.findall('href="/sf/go/(\w+\d+)[^>]*">(.*?)</a>', content)
        # mark all intra-project object first
        for objId, text in objId_list:
            if objId in self.old_new:
                if text.find(objId) >= 0:
                    new_text = text.replace(objId, 'OLD:%s' % objId)
                    content = content.replace('href="/sf/go/%s[^>]*">%s</a>' % (objId, text),
                                              'href="/sf/go/OLD:%s[^>]*">%s</a>' % (objId, new_text))
                else:
                    content = content.replace('href="/sf/go/%s"' % objId, 
                                              'href="/sf/go/OLD:%s"' % objId)
                found.append(objId)
        if len(found) > 0:
            # now substitue old objId with new ID
            for objId in found:
                content = content.replace('OLD:%s' % objId, 
                                          '%s' % self.old_new[objId])
        # update image urls: src="/sf/page/do/viewImage/projects.<name>/page/xxx.jpg
        oName = self.orig_projectName
        nName = self.session.Ctx['project'].name
        content = content.replace('src="/sf/page/do/viewImage/projects.%s/page/' % oName,
                                  'src="/sf/page/do/viewImage/projects.%s/page/' % nName)
        return content
        
    def add_subfolders(self, child, newParentId, folderType):
        """Add child node to specified parent node"""

        subfolder = self.nodes[child]
        parentId = subfolder.find('parentFolderId').text
        title = subfolder.find('title').text
        path = subfolder.find('path').text
        description = subfolder.find('description').text
        if parentId not in self.old_new:
            self.old_new[parentId] = newParentId
        if description is None:
            description = ' '   # FIX: allow empty value
        self.log.verbose("  created subfolder '%s'" % title)
        if parentId not in self.old_new:
            self.log.error("parentId not in old_new association map, '%s'" % parentId)
            self.log.error("you should not reach here!")
            return
        if folderType == "PlanningFolder":
            cmd = 'createPlanningFolder'
            s = subfolder.find('status')
            if s is not None:
                status = s.text
            else:
                status = None
            r = subfolder.find('releaseId')
            if r is not None:
                old_releaseId = r.text
                if old_releaseId is not None and old_releaseId in self.old_new:
                    releaseId = self.old_new[old_releaseId]
                elif old_releaseId is None:
                    releaseId = None
                else:
                    self.log.warn("releaseId not in old_new association map, '%s'" % old_releaseId)
                    self.log.warn("setting value for 'releaeId' field to 'None'")
                    releaseId = None
            else:
                # <releaseId> node is not found in data
                releaseId = None
            c = subfolder.find('capacity')
            if c is not None and c.text is not None:
                # found <capacity> node, and it has a value
                capacity = int(c.text)
            else:
                capacity = None
            argvs = [self.old_new[parentId], title, description, 'none', 'none',
                     status, capacity, releaseId]
        elif folderType == "DocumentFolder":
            cmd = 'createDocumentFolder'
            argvs = [self.old_new[parentId], title, description]

        new_subfolder = ApiMethods[cmd](self.session, cmd, argvs).export_xml()
        if self.session.error > 0:
            self.log.error("Failed to create subfolder for '%s' project" % self.projectId)
            self.logoff_exit(1)

        new_subfolderId = new_subfolder.find('id').text
        new_path = new_subfolder.find('path').text
        self.path_map[path] = new_path
        self.log.debug2("    old docFolder = '%s'" % path)
        self.log.debug2("    new docFolder = '%s'" % new_path)
        self.log.debug5("  cmd (argvs): %s (%s)" % (cmd, argvs))
        self.old_new[child] = new_subfolderId
        self.new_old[new_subfolderId] = child
        self.nnodes[new_subfolderId] = new_subfolder
        self.subfolder_count += 1

        if child in self.map and self.map[child] is not None:
            for xid in self.map[child]:
                if  child in self.old_new:
                    self.add_subfolders(xid, self.old_new[child], folderType)
                else:
                    self.log.error("parentId not found in old_new association map, '%s'" % child)
                    self.log.error("You should not get this error :( !")

    def load_RbacApp(self, roles):
        if roles is None:
            self.log.warn("There are no project roles to be created.")
            self.log.info("")
            return
        self.log.info("Creating project roles ...")
        n = 0
        for role in roles:
            preventInheritance = requestable = autoGrant = False
            Id = role.find('id').text
            description = role.find('description').text
            title = role.find('title').text
            clusters = role.find('Clusters')
            if (role.find('preventInheritance') is not None):
                preventInheritance = DataUtils.to_boolean(role.find('preventInheritance').text)
            if (role.find('requestable') is not None):
                requestable = DataUtils.to_boolean(role.find('preventInheritance').text)
            if (role.find('autoGrant') is not None):
                autoGrant = DataUtils.to_boolean(role.find('autoGrant').text)
            cmd = 'createRole'
            argvs = [self.projectId, title, description, preventInheritance, requestable, autoGrant ]
            new_role = ApiMethods[cmd](self.session, cmd, argvs).export_xml()
            if self.session.error > 0:
                self.log.error("Can't create new role '%s'" % title)
                self.logoff_exit(1)
            else:
                n += 1
                self.log.verbose("  created role '%s'" % title)
            new_id = new_role.find('id').text
            self.old_new[Id] = new_id
            self.add_clusters(new_id, clusters)
        self.log.info("Finished creating all project roles successfully, total = %s." % n)
        self.log.info("")

    def add_clusters(self, roleId, clusters):
        cmd = 'addCluster'
        for cluster in clusters:
            folderId = cluster.find('folderId').text
            if folderId and folderId in self.old_new:
                new_folderId = self.old_new[folderId]
            else:
                if folderId :
                    self.log.warn("      folder '%s' not found on this site" % folderId)
                    self.log.warn("      it may be external to the project")
                new_folderId = None

            name = cluster.find('operationClusterName').text
            argvs = [roleId, name, new_folderId]
            ApiMethods[cmd](self.session, cmd, argvs).run()
            if self.session.error > 0:
                self.log.error("    Can't create new cluster '%s' for role '%s'" % (name, roleId))
                self.logoff_exit(1)
            else:
                self.log.debug("    created operationCluster '%s'" % name)

    def load_FrsApp(self, packages):
        if packages is None:
            self.log.info("There are no project file releases to be created.")
            self.log.info("")
            return
        self.log.info("Creating project file releases ...")
        pkg_total = release_total = 0
        for package in packages:
            Id = package.find('id').text
            description = package.find('description').text
            title = package.find('title').text
            isPublished = package.find('isPublished').text
            releases = package.find('Releases')
            if isPublished:
                isPublished = isPublished.lower() == 'true'
            else:
                isPublished = False

            cmd = 'createPackage'
            argvs = [self.projectId, title, description, isPublished]
            new_package = ApiMethods[cmd](self.session, cmd, argvs).export_xml()
            if self.session.error > 0:
                self.log.error("Can't create new package '%s'" % title)
                self.logoff_exit(1)
            else:
                pkg_total += 1
                self.log.verbose("  created package '%s'" % title)
            new_pkgId = new_package.find('id').text
            self.old_new[Id] = new_pkgId
            release_total += self.add_releases(new_pkgId, title, releases)
        self.log.info("Finished creating total of %s releases for %s packages." % \
                       (release_total, pkg_total))
        self.log.info("")

    def add_releases(self, pkgId, pkgTitle, releases):
        cmd = 'createRelease'
        releaseId = new_rId = None
        total = 0
        for release in releases:
            releaseId = release.find('id').text
            title = release.find('title').text
            status = release.find('status').text
            maturity = release.find('maturity').text
            description = release.find('description').text
            argvs = [pkgId, title, description, status, maturity]
            new_release = ApiMethods[cmd](self.session, cmd, argvs).export_xml()
            if self.session.error > 0:
                self.log.error("    Can't create release '%s' for package '%s'" % (title, pkgTitle))
                return
            else:
                self.log.debug("    created release '%s' for package '%s'" % (title, pkgTitle))
                new_rId = new_release.find('id').text
                self.old_new[releaseId] = new_rId
                total += 1
        return total
                
    def create_Template(self):
        self.log.info("Creating template from newly created project")
        projectId = self.projectId
        title = self.session.Ctx['project'].title
        description = self.session.Ctx['project'].description
        cmd = 'createOrReplaceProjectTemplate'
        argvs = [projectId, title, description, ['0'], 'none']
        ApiMethods[cmd](self.session, cmd, argvs).export_xml()
        if self.session.error > 0:
            self.log.error("Can't create template from project '%s'" % projectId)
            self.logoff_exit(1)
        else:
            self.log.verbose("  created project template '%s'" % title)
        self.log.info("Finished creating template from new project successfully.")
        self.log.info("")

    def logoff_exit(self, exitCode):
        if self.session.session_id:
            ApiMethods['logoff'](self.session, 'logoff', [self.session.user, self.session.session_id]).run()
            sys.exit(exitCode)
