
# __revision__  = $LastChangedRevision$
# __Copyright__ = Copyright (c) 2007 - 2009, CollabNet, Inc. All rights reserved.

import os
import sys
from lxml import etree
from ZSI import FaultException

BASEDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
sys.path.append(os.path.join(BASEDIR, "TeamForge"))

from SimpleLog import SimpleLog
from Soap60.PlanningApp import PlanningAppSoapService_services as PlanningServices
import DataUtils
import CTFApp

# bind to a dummy url in order to get all available services
Services = [ m for m in dir(PlanningServices.PlanningAppSoapBindingSOAP('')) \
               if not (m.startswith('_') or m.startswith('binding'))]

SoapApp = 'PlanningApp'

class PlanningApp(object):
    def __init__(self, session, cmd, args=None):
        self.session = session
        self.session.error = 0
        self.loglevel = self.session.loglevel
        self.cmd = cmd
        self.args = args
        #self.mc = self.session.mc
        self.results = None
        self.project_id = None
        self.log = SimpleLog(self.loglevel)

        if self.session.base_url is None:
            self.log.error("TeamForge server is not specified.")
            self.session.error = 1
            return
        else:
            self.port = self.get_service_binding()
        if self.port is None:
            self.log.error("SoapApp URL is not set for '%s'" % SoapApp)
            self.session.error = 1
            return

        self.request = self.get_request()
        self.params, self.ptypes = DataUtils.get_params_types(self.request)
        self.tagname = DataUtils.get_tagname(self.cmd)

        if self.loglevel == "DEBUG5":
            self.log.debug5("")
            self.log.debug5("request parameters for cmd '%s':" % self.cmd)
            for idx in range(len(self.params)):
                self.log.debug5("     request.%s (%s) " % (self.params[idx],
                                                   self.ptypes[idx]))


        if not self.session.help:
            self.action()

    def get_service_binding(self):
        url = self.session.get_app_url(SoapApp)
        if url:
            port = PlanningServices.PlanningAppSoapBindingSOAP(url, **self.session.kw_soapopt)
        else:
            return None
        return port

    def get_request(self):
        req_name = self.cmd + "Request"
        return getattr(PlanningServices, req_name)()

    def usage(self, errmsg=None):
        if errmsg:
            self.log.error("%s" % errmsg)
        self.log.info("%s %s" % (self.cmd, ' '.join(self.params[1:])))

    def validate(self):
        need_projectId = [ 'getPlanningStatusValues', 'setPlanningStatusValues']
        need_planningId = ['deletePlanningFolder', 'getPlanningFolderData', 
                           'getPlanningFolderSummary', 'getRankedArtifactList',
                           'movePlanningFolder', ]

        # session_id is always the first argument for all methods
        if not self.session.session_id:
            self.log.error("active session not found, please login first")
            return False
        else:
            self.args.insert(0, self.session.session_id)

        # extract project_id from args or from context
        if len(self.args) == len(self.params) :
            self.project_id = self.args[1]
        elif self.session.Ctx.has_key('project') and  \
             self.session.Ctx['project'] is not None :
            self.project_id = self.session.Ctx['project'].id
        else:
            self.log.debug("%s: could not get projectId" % self.cmd)

        # if project context is set and user didnot provide projectId, 
        # use projectId from project context
        if self.session.Ctx.has_key('project') and  \
           self.session.Ctx['project'] is not None and \
           self.cmd in need_projectId and len(self.args) == len(self.params) - 1:
            self.args.insert(1, self.session.Ctx['project'].id)

        # if planning context is set and user didnot provide planningId, 
        # use planningId from project context
        if self.session.Ctx.has_key('planningfolder') and  \
           self.session.Ctx['planningfolder'] is not None and \
           self.cmd in need_projectId and len(self.args) == len(self.params) - 1:
            self.args.insert(1, self.session.Ctx['planningfolder'].id)

        if len(self.args) < len(self.params) :
            self.usage("syntax error: missing argument")
            return False
        elif len(self.args) > len(self.params) :
            self.usage("syntax error: too many argument")
            return False
        else:

            num_args = len(self.args)
            for idx in range(len(self.params)):
                method = "set_element_" + self.params[idx]
                if idx >= num_args:
                    val = None
                elif self.ptypes[idx] == 'boolean':
                    val = DataUtils.to_boolean(self.args[idx])
                elif self.ptypes[idx] == 'int':
                    val = DataUtils.to_int(self.args[idx])
                elif self.ptypes[idx] == 'long':
                    val = DataUtils.to_long(self.args[idx])
                elif self.ptypes[idx] == 'string':
                    val = DataUtils.to_string(self.args[idx])
                elif self.ptypes[idx] == 'dateTime':
                    val = DataUtils.to_dateTime(self.args[idx])
                elif self.ptypes[idx] == 'ArrayOf_xsd_string':
                    val = DataUtils.to_array(self.args[idx])
                elif self.ptypes[idx] == 'ArrayOf_tns1_PlanningStatusValueSoapDO':
                    val = self.update_StatusValues(self.args[idx])
                else:
                    val = self.args[idx]
                if val is not None:
                    getattr(self.request, method)(val)
    
                if self.loglevel == 'DEBUG5':
                    getter = method.replace('set', 'get')
                    self.log.debug5("Planning.validate request.%s => %s (%s)" % \
                          (getter, repr(getattr(self.request, getter)()), self.ptypes[idx]))
        return True

    def action(self):
        if not self.validate():
            return
        try:
            response = getattr(self.port, self.cmd)(self.request)
        except FaultException, fe:
            fc = fe.fault.code[1]
            fs = fe.fault.string
            if fs.find('No such operation') >= 0:
                self.log.warn("%s" % fs)
                self.log.warn("Some operations failed")
                self.log.warn("Please check ApiVersion supported by server")
            else:
                self.log.debug5("cmd failed: '%s %s'" % (self.cmd, self.args))
                self.log.error("error code    = %s" % fc)
                self.log.error("error message = %s" % fs)
            self.session.error = 1
            return
        except TypeError, te:
            major, minor = DataUtils.get_PythonVersion()[0:2]
            if major > 2 or major == 2 and minor > 4:
                fs = repr(te)
            else:
                fs = te
            self.log.debug5("cmd failed: '%s %s'" % (self.cmd, self.args))
            self.log.error("error message = %s" % fs)
            self.session.error = 1
            return
        if response:
            getReturn = 'get_element_' + self.cmd + 'Return'
            if getReturn in dir(response):
                try:
                    self.results = getattr(response, getReturn)()
                except Exception, ee:
                    self.log.error("error message: %s" % repr(ee))

    def run(self):
        if self.results is None:
            return
        elif isinstance(self.results, bool):
            if self.results:
                print >> sys.stderr, "True"
            else:
                print >> sys.stderr, "False"
            return
        elif self.cmd.startswith('createPlanningFolder'):
            pf_data = self.export_xml()
            if pf_data is not None:
                self.session.Ctx['planningfolder'] = CTFApp.PlanningFolder(pf_data)
        else:
            xml = self.export_xml()
            if xml is None:
                return
            elif self.session.format == 'xml':
                print etree.tostring(xml, encoding='UTF-8', pretty_print=True)
            elif self.session.format == 'table':
                DataUtils.export_table(xml)

    def export_xml(self):
        return DataUtils.export_xml(self.results, root_name=self.tagname)

    def export_dict(self):
        return DataUtils.export_dict(self.results)

    def export_data(self):
        return self.results

    def _get_current_status_values(self):
        new_cmd = 'getPlanningStatusValues'
        return PlanningApp(self.session, new_cmd, [self.project_id]).export_xml()

    def update_StatusValues(self, values):
        """values is an list of tuples: [(isDefault, value, valueClass), ...]
           if an existing value is specified, it is updated;
           if an existing value is not present, it is deleted;
           if a new value is specified, it is added.
        """
        current_sv_dict = {}

        if isinstance(values, list):
            data = values
        elif isinstance(values, str):
            data = eval(values)
            if not isinstance(data, list):
                self.log.error("setPlanningStatusValues data format error")
                sys.exit(1)

        old_sv = self._get_current_status_values()
        for item in old_sv:
            k = item.find('value').text
            current_sv_dict[k] = item

        new_values = []
        for item in data:
            if not isinstance(item, tuple):
                self.log.error("setPlanningStatusValues value format error")
                sys.exit(1)
            is_default, value, valueClass = item
            isDefault = DataUtils.to_boolean(is_default)
            sv = PlanningStatusValues()
            setattr(sv, '_isDefault', isDefault)
            setattr(sv, '_value', value)
            setattr(sv, '_valueClass', valueClass)
            if value in current_sv_dict:
                # updating exisiting value
                setattr(sv, '_id', current_sv_dict[value].find('id').text)
            new_values.append(sv)
        return new_values

# Due to limitation in ZSI(version 2.0.4) library, factory methods for creating
# some Complextype SOAP objects were not generated by wsdl2py tool. Therefore
# follwoing definitions
# had to be made here to facilitate use of various SOAP calls.

from ZSI.TCcompound import Struct
import ZSI
from ZSI.generate.pyclass import pyclass_type

_PlanningStatusValuesTypecode = Struct(\
    pname=("http://schema.open.collab.net/sfee50/soap60/service","PlanningStatusValues"),
    ofwhat=[ZSI.TC.String(pname="id", aname="_id", minOccurs=1, maxOccurs=1, nillable=True,
                          typed=False, encoded=None),
            ZSI.TC.Boolean(pname="isDefault", aname="_isDefault", minOccurs=1, maxOccurs=1,
                          nillable=False, typed=False, encoded=None),
            ZSI.TC.String(pname="value", aname="_value", minOccurs=1, maxOccurs=1, nillable=True,
                          typed=False, encoded=None),
            ZSI.TC.String(pname="valueClass", aname="_valueClass", minOccurs=1, maxOccurs=1,
                          nillable=True, typed=False, encoded=None)],
    pyclass=None,
    encoded="http://schema.open.collab.net/sfee50/soap60/service")

class PlanningStatusValues:
    typecode = _PlanningStatusValuesTypecode
    __metaclass__ = pyclass_type
    def __init__(self):
        self._id = None
        self._isDefault = None
        self._value = None
        self._valueClass = None
        return
PlanningStatusValues.typecode.pyclass = PlanningStatusValues

