
# __revision__  = $LastChangedRevision$
# __Copyright__ = Copyright (c) 2007 - 2009, CollabNet, Inc. All rights reserved.

import os
import sys
from lxml import etree
from ZSI import FaultException

BASEDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
sys.path.append(os.path.join(BASEDIR, "TeamForge"))

from Soap60.TrackerApp import TrackerAppSoapService_services as TrackerServices
import CTFApp
import DataUtils
from SimpleLog import SimpleLog

# bind to a dummy url in order to get all available services
Services = [ m for m in dir(TrackerServices.TrackerAppSoapBindingSOAP('')) \
               if not (m.startswith('_') or m.startswith('binding'))]

SoapApp = 'TrackerApp'

class TrackerApp(object):
    def __init__(self, session, cmd, args=None):
        self.session = session
        self.session.error = 0
        self.loglevel = self.session.loglevel
        self.cmd = cmd
        self.args = args
        #self.mc = self.session.mc
        self.results = None
        self.log = SimpleLog(self.session.loglevel)

        if self.session.base_url is None:
            self.log.error("TeamForge server is not specified.")
            self.session.error = 1
            return
        else:
            self.port = self.get_service_binding()
        if self.port is None:
            self.log.error("SoapApp URL is not set for '%s'" % SoapApp)
            self.session.error = 1
            return

        self.request = self.get_request()
        self.params, self.ptypes = DataUtils.get_params_types(self.request)
        self.tagname = DataUtils.get_tagname(self.cmd)

        if self.loglevel == "DEBUG5":
            self.log.debug5("")
            self.log.debug5("request parameters for cmd '%s':" % self.cmd)
            for idx in range(len(self.params)):
                self.log.debug5("    request.%s (%s) " % (self.params[idx],
                                                   self.ptypes[idx]))
        if not self.session.help:
            self.action()

    def get_service_binding(self):
        url = self.session.get_app_url(SoapApp)
        if url:
            port = TrackerServices.TrackerAppSoapBindingSOAP(url, **self.session.kw_soapopt)
        else:
            return None
        return port

    def get_request(self):
        req_name = self.cmd + "Request"
        return getattr(TrackerServices, req_name)()

    def usage(self, errmsg=None):
        if errmsg:
            self.log.error("%s" % errmsg)
        self.log.info("Usage: %s %s" % (self.cmd, ' '.join(self.params[1:])))

    def validate(self):
        need_projectId = ['createTracker', 'getTrackerList']
        need_trackerId = ['addDateField', 'addMultiSelectField', 'addSingleSelectField',
                          'addTextField', 'addUserSelectField', 'createArtifact',
                          'getAllowedWorkflowTransitionList', 'getArtifactDetailList',
                          'getFields', 'getTrackerData']

        # session_id is always the first argument for all methods
        if not self.session.session_id:
            self.log.error("active session not found, please login first")
            return False
        else:
            self.args.insert(0, self.session.session_id)

        # if project context is set and user didnot provide projectId, 
        # use projectId from project context
        if self.session.Ctx.has_key('project') and  \
           self.session.Ctx['project'] is not None and \
           self.cmd in need_projectId and len(self.args) < len(self.params) :
            self.args.insert(1, self.session.Ctx['project'].id)

        # if tracker context is set and user didnot provide tracekrId, 
        # use tracekrId from tracekr context
        if self.session.Ctx.has_key('tracker') and  \
           self.session.Ctx['tracker'] is not None and \
           self.cmd in need_trackerId and len(self.args) < len(self.params) :
            self.args.insert(1, self.session.Ctx['tracker'].id)

        if len(self.args) < len(self.params) :
            self.usage("syntax error: missing argument")
            return False
        elif len(self.args) > len(self.params) :
            self.usage("syntax error: too many argument")
            return False
        else:
            for idx in range(len(self.params)):
                method = "set_element_" + self.params[idx]

                if self.ptypes[idx] == 'boolean':
                    val = DataUtils.to_boolean(self.args[idx])
                elif self.ptypes[idx] == 'int':
                    val = DataUtils.to_int(self.args[idx])
                elif self.ptypes[idx] == 'long':
                    val = DataUtils.to_long(self.args[idx])
                elif self.ptypes[idx] == 'string':
                    val = DataUtils.to_string(self.args[idx])
                elif self.ptypes[idx] == 'ArrayOf_xsd_string':
                    val = DataUtils.to_array(self.args[idx])
                elif self.ptypes[idx] == 'SoapFieldValues':
                    val = self.create_flexFields(self.args[idx])
                elif self.ptypes[idx] == 'TrackerSoapDO':
                    val = self.create_trackerData(self.args[idx])
                elif self.ptypes[idx] == 'TrackerFieldSoapDO':
                    val = self.create_fieldData(self.args[idx])
                elif self.ptypes[idx] == 'ArtifactSoapDO':
                    val = self.create_artifactData(self.args[idx], objType='ArtifactSoapDO')
                #TODO: add methods for these:
                #       ArrayOf_tns1_SoapFilter
                #       ArrayOf_tns1_SoapSortKey
                else:
                    val = self.args[idx]
                if val is not None:
                    getattr(self.request, method)(val)

                if self.loglevel == 'DEBUG5':
                    getter = "get_element_" + self.params[idx]
                    self.log.debug5("    request.%s => %s" % \
                          (getter, getattr(self.request, getter)()))
        return True

    def action(self):
        if not self.validate():
            return
        try:
            response = getattr(self.port, self.cmd)(self.request)
        except FaultException, fe:
            fc = fe.fault.code[1]
            fs = fe.fault.string
            self.log.debug5("cmd failed: '%s %s'" % (self.cmd, self.args))
            self.log.error("error code    = %s" % fc)
            self.log.error("error message = %s" % fs)
            self.session.error = 1
            return
        except TypeError, te:
            major, minor = DataUtils.get_PythonVersion()[0:2]
            if major > 2 or major == 2 and minor > 4:
                fs = repr(te)
            else:
                fs = te
            self.log.debug5("cmd failed: '%s %s'" % (self.cmd, self.args))
            self.log.error("TypeError, error message = %s" % fs)
            self.session.error = 1
            return

        if response:
            getReturn = 'get_element_' + self.cmd + 'Return'
            if getReturn in dir(response):
                try:
                    self.results = getattr(response, getReturn)()
                except Exception, ee:
                    self.log.error("error message: %s" % repr(ee))

    def run(self):
        if self.results is None:
            return
        elif isinstance(self.results, bool):
            if self.results:
                print >> sys.stderr, "True"
            else:
                print >> sys.stderr, "False"
            return
        elif self.cmd.startswith('create'):
            if self.cmd == "createTracker":
                tracker_data = self.export_xml()
                if tracker_data is not None:
                    self.session.Ctx['tracker'] = CTFApp.Tracker(tracker_data)
        else:
            xml = self.export_xml()
            if xml is None:
                return
            elif self.session.format == 'xml':
                print etree.tostring(xml, encoding='UTF-8', pretty_print=True)
            elif self.session.format == 'table' and not xml.text:
                DataUtils.export_table(xml)

    def export_xml(self):
        return DataUtils.export_xml(self.results, root_name=self.tagname)

    def export_dict(self):
        return DataUtils.export_dict(self.results)

    def export_data(self):
        return self.results

    def create_trackerData(self, val):
        pass

    def create_artifactData(self, val, objType=None):
        """Except for id field, only modified field need to be provided"""

        if objType is None or val is None or val.lower() == 'none' or val.lower() == 'nil':
            return []
        data = eval(val)
        if not isinstance(data, dict):
            self.log.error("input artifactData format error")
            sys.exit(1)
 
        if 'id' in data:
            xid = data['id']
        else:
            self.log.error("input data must contain id field")
            sys.exit(1)
        if xid in self.session.artifacts:
            artfData = self.session.artifacts[xid]
        else:
            self.update_ArtifactMap(xid, objType)
            if not xid in self.session.artifacts:
                self.log.error("TrackerApp: artifact data not found in session"
                               " cache for '%s'" % xid)
                self.log.error("this situation should not happen")
                return 
            else:
                artfData = self.session.artifacts[xid]
        new_artfData = self.update_artifactData(artfData)

        for key in data.keys():
            if key in ['lastModifiedDate']:
                continue
            if key == 'flexFields':
                value = self.create_flexFields(data[key])
            elif key in ['createDate', 'closeDate']:
                v = data[key]
                if v:
                    value = MyDateTuple(DataUtils.to_dateTime(v))
                else:
                    value = None
            elif key == 'autosumming':
                value = DataUtils.to_boolean(data[key])
            elif key.endswith('Effort') or key.endswith('Hours') or \
                 key == 'priority' or key == 'version':
                value = int(data[key])
            else:
                value = data[key]
            aname = key[0].upper() + key[1:]
            setattr(new_artfData, aname, value)

        if self.loglevel == "DEBUG5":
            getter_methods = [m1 for m1 in dir(artfData) \
                                  if m1.startswith('get_element_')]
            for method in getter_methods:
                self.log.debug5("    artfData.%s => %s" % \
                                (method, getattr(new_artfData, method)()))
        return new_artfData

    def create_flexFields(self, val):
        """argument val is a str representation of a list containing three lists
           of equal length
        """
        flexFields = self.request.new_flexFields()

        if val is None or val.lower() == 'none' or val.lower() == 'nil':
            return []
        data = eval(val)
        if not isinstance(data, list):
            self.log.error("input flexfields data format error")
            sys.exit(1)
        elif data[0] is None or data[1] is None or data[2] is None:
            return flexFields
        else:
            if isinstance(data[0], list):
                l0 = len(data[0])
                d0 = data[0]
            elif isinstance(data[0], str):
                l0 = len(eval(data[0]))
                d0 = eval(data[0])

            if isinstance(data[1], list):
                l1 = len(data[1])
                d1 = data[1]
            elif isinstance(data[1], str):
                l1 = len(eval(data[1])) 
                d1 = eval(data[1])

            if isinstance(data[2], list):
                l2 = len(data[2])
                d2 = data[2]
            elif isinstance(data[2], str):
                l2 = len(eval(data[2])) 
                d2 = eval(data[2])

            if l0 != l1 or l0 != l2:
                self.log.error("flexfields components have variable sizes:")
                self.log.error("names(%s), types(%s), values(%s)" % (l0, l1, l2))
                sys.exit(1)

            # d0, d1, d2 all must be list
            for i in range(len(d1)):
                if d1[i] == 'Date':
                    if d2[i]:
                        d2[i] = MyDateTuple(d2[i])
        flexFields.set_element_names(d0)
        flexFields.set_element_types(d1)
        flexFields.set_element_values(d2)
        if self.loglevel == "DEBUG5":
            getter_methods = [m1 for m1 in dir(flexFields) \
                                  if m1.startswith('get_element_')]
            for method in getter_methods:
                self.log.debug5("    flexFields.%s => %s" % \
                                (method, getattr(flexFields, method)()))
        return flexFields

    def create_fieldData(self, val):
        """argument val is a str representation of a dictionary contains values for 
        fields to be updated."""
        if val is None or val.lower() == 'none' or val.lower() == 'nil':
            return None
        data = eval(val)
        if not isinstance(data, dict):
            self.log.error("setField data format error")
            sys.exit(1)

        fieldData = self.request.new_fieldData()

        xid = data['id']
        if xid in self.session.trackerFields:
            current_fv = self.session.trackerFields[xid].find('fieldValues')
        else:
            self.update_FieldMap()
            if not xid in self.session.trackerFields:
                self.log.error("TrackerApp: trackerField data not found in session"
                               " cache for '%s'" % xid)
                self.log.error("this situation should not happen")
                return fieldData
            else:
                current_fv = self.session.trackerFields[xid].find('fieldValues')

        if 'fieldValues' in data:
            new_fieldValues = self.update_FieldValues(data['fieldValues'], current_fv)
        else:
            new_fieldValues = []
     
        for key in data.keys():
            name = key[0].upper() + key[1:]
            if name in ['Disabled', 'HiddenOnCreate', 'Required']:
                if data[key] is None:
                    val = False
                elif isinstance(data[key], bool):
                    val = data[key]
                elif isinstance(data[key], str):
                    val = data[key].lower() == 'true'
                setattr(fieldData, name, val)
            elif name in ['DisplayLines', 'DisplaySize', 'Version']:
                setattr(fieldData, name, int(data[key]))
            elif name.startswith('CreatedDate') or name.startswith('LastModifiedDate'):
                setattr(fieldData, name, DataUtils.to_dateTime(data[key]))
            elif name == 'FieldValues':
                setattr(fieldData, name, new_fieldValues)
            else:
                setattr(fieldData, name, data[key])

        if self.loglevel == "DEBUG5":
            getter_methods = [m1 for m1 in dir(fieldData) if m1.startswith('get_element_')]
            for method in getter_methods:
                self.log.debug5("    fieldData.%s => %s" % (method, getattr(fieldData, method)()))

        return fieldData

    def update_artifactData(self, artf_xml):
        Data = self.request.new_artifactData()
        new_flexfields = Data.new_flexFields()

        for child in artf_xml.iter():
            tag = child.tag
            if tag in ['lastModifiedDate']:
                continue
            if tag == 'autosumming':
                val = DataUtils.to_boolean(child.text)
            elif tag in ['createdDate', 'closeDate']:
                v = child.text
                if v:
                    val = MyDateTuple(DataUtils.to_dateTime(v))
                else:
                    val = None
            elif tag == 'flexFields':
                val = self.update_flexFields(child, new_flexfields)
            elif tag.endswith('Effort') or tag.endswith('Hours') or \
                 tag == 'priority' or tag == 'version':
                val = int(child.text)
            else:
                val = child.text
            name = tag[0].upper() + tag[1:]
            if val is not None:
                setattr(Data, name, val)
        return Data

    def update_flexFields(self, old_FF, new_FF):
        """old_FF is etree.Element tree object"""

        names = old_FF.find('names').text
        types = old_FF.find('types').text
        values = old_FF.find('values').text

        #convert to list
        if names:
            names = eval(names)
            types = eval(types)
            values = eval(values)
            for i in range(len(types)):
                if types[i] == 'Date':
                    if values[i]:
                        values[i] = MyDateTuple(values[i])
        new_FF.set_element_names(names)
        new_FF.set_element_types(types)
        new_FF.set_element_values(values)
        return new_FF

    def update_FieldValues(self, values, current_fv):
        """values argument is updatd values, it should be a list of tuples: 
                  [(value, valueClass, isDefault),]
           current_fv is a TrackerFieldValueSoapDO object from the tracker field whose
                  values are to be updated
        """
        newFV = []
        fvmap = {}
        
        if len(current_fv) > 0:
            for c_fv in current_fv.iter('fieldValue'):
                if len(c_fv) > 0:
                    fid = c_fv.find('id').text
                    fval = c_fv.find('value').text
                    fvmap[fval] = fid
       
        for val, valCls, isDefault in values:
            # this is a new value
            fv = TrackerFieldValues()
            if val in fvmap:
                fv._id = fvmap[val]
            fv._value = val
            fv._valueClass = valCls
            if isinstance(isDefault, str):
                fv._isDefault = isDefault.lower() == 'true'
            elif isinstance(isDefault, bool):
                fv._isDefault = isDefault
            elif isDefault is None:
                fv._isDefault = False
            newFV.append(fv)

            if self.loglevel == "DEBUG5":
                getter_methods = [m1 for m1 in dir(fv) if m1.startswith('get_element_')]
                for method in getter_methods:
                    self.log.debug5("    new fieldValues.%s => %s" % \
                                     (method, getattr(fv, method)()))
        return newFV

    def update_FieldMap(self):
        if 'tracker' in self.session.Ctx:
            trackerId = self.session.Ctx['tracker'].id
        else:
            self.log.error("(update_FieldMap): tracker context is not set")
            sys.exit(1)

        fields = TrackerApp(self.session, 'getFields', [trackerId]).export_xml()
        if self.session.error > 0:
            self.log.error("(update_FieldMap): can't retrieve tracker fields data")
            sys.exit(1)
        for f in fields.iter('Field'):
            xid = f.find('id').text
            self.session.trackerFields[xid] = f

    def update_ArtifactMap(self, artfId, objType):
        if objType == 'ArtifactSoapDO':
            cmd = 'getArtifactData'
        else:
            self.log.error("Unknown ArtifactSoapDO version '%s', exiting" % objType)
            sys.exit(1)
        artfData = TrackerApp(self.session, cmd, [artfId]).export_xml()
        if self.session.error > 0:
            self.log.error("(update_FieldMap): can't retrieve tracker fields data")
            sys.exit(1)
        self.session.artifacts[artfId] = artfData

# Due to limitation in ZSI(version 2.0.4) library, factory methods for creating some
# Complextype SOAP objects were not generated by wsdl2py tool. Therefore follwoing definitions
# had to be made here to facilitate use of various SOAP calls.

from ZSI.TCcompound import Struct
import ZSI

from ZSI.generate.pyclass import pyclass_type

class MyDateTuple(tuple):
    typecode = ZSI.TCtimes.gDateTime()

_TrackerFieldValuesTypecode = Struct(\
    pname=("http://schema.open.collab.net/sfee50/soap60/service","TrackerFieldValues"), 
    ofwhat=[ZSI.TC.String(pname="id", aname="_id", minOccurs=1, maxOccurs=1, nillable=True, 
                          typed=False, encoded=None), 
            ZSI.TC.Boolean(pname="isDefault", aname="_isDefault", minOccurs=1, maxOccurs=1, 
                          nillable=False, typed=False, encoded=None), 
            ZSI.TC.String(pname="value", aname="_value", minOccurs=1, maxOccurs=1, nillable=True, 
                          typed=False, encoded=None), 
            ZSI.TC.String(pname="valueClass", aname="_valueClass", minOccurs=1, maxOccurs=1, 
                          nillable=True, typed=False, encoded=None)], 
    pyclass=None, 
    encoded="http://schema.open.collab.net/sfee50/soap60/service")

class TrackerFieldValues:
    typecode = _TrackerFieldValuesTypecode
    __metaclass__ = pyclass_type
    def __init__(self):
        self._id = None
        self._isDefault = None
        self._value = None
        self._valueClass = None
        return
TrackerFieldValues.typecode.pyclass = TrackerFieldValues
