
from ImportModule import ApiMethods
from SimpleLog import SimpleLog
import DataUtils

# __revision__  = $LastChangedRevision$
# __Copyright__ = Copyright (c) 2007 - 2009, CollabNet, Inc. All rights reserved.

commands = ['pageTreeView', 'docfolderTreeView', 'planningTreeView']

class TreeView(object):
    def __init__(self, session, args=None):
        self.session = session
        self.args = args
        self.projectId = None
        self.children = {}
        self.nodes = {}
        self.format = self.session.format
        self.log = SimpleLog(self.session.loglevel)

    def usage(self, errmsg=None):
        """Should be overridden"""
        pass

    def validate(self):
        if not self.session.session_id:
            self.log.warn("no active session, please log in first")
            return False
        if self.args and len(self.args) == 1:
            self.projectId = self.args[0]
        elif 'project' in self.session.Ctx and self.session.Ctx['project'] is not None:
            self.projectId = self.session.Ctx['project'].id
        else:
            self.usage("projectId is not set")
            return False

        return True
        
    def action(self, app, cmd):
        if not self.validate():
            return
        indent = ''
        root = None
        variables = [self.projectId]
        if app in ['PlanningFolder', 'DocumentFolder'] :
            variables.append('true')
        data = ApiMethods[cmd](self.session, cmd, variables).export_xml()
        if data is not None:
            root, self.children, self.nodes = DataUtils.find_children(data)
        else:
            self.log.error("%s not found for project '%s'" % (app, self.projectId))
            return

        self.log.info("%s Tree View:" % app)
        self.log.info("%s" % root)
        DataUtils.print_treeview(root, self. children, self.nodes, indent)

    def run(self):
        """To be implemented"""
        pass

class planningTreeView(TreeView):
    def __init__(self, session, args=None):
        TreeView.__init__(self, session, args)

    def usage(self, errmsg=None):
        if errmsg:
            self.log.error('%s' % errmsg)
        self.log.info("Print out project planning folders in a tree view.")
        self.log.info("Usage: planningTreeView projectId")
  
    def run(self):
        self.action('PlanningFolder', 'getPlanningFolderList')

class pageTreeView(TreeView):
    def __init__(self, session, args=None):
        TreeView.__init__(self, session, args)

    def usage(self, errmsg=None):
        if errmsg:
            self.log.error('%s' % errmsg)
        self.log.info("Print out project pages in a tree view.")
        self.log.info("Usage: pageTreeView projectId")
  
    def run(self):
        self.action('ProjectPages', 'getPageList')

class docfolderTreeView(TreeView):
    def __init__(self, session, args=None):
        TreeView.__init__(self, session, args)

    def usage(self, errmsg=None):
        if errmsg:
            self.log.error('%s' % errmsg)
        self.log.info("Print out project document folders in a tree view.")
        self.log.info("Usage: docFolderTreeView projectId")
  
    def run(self):
        self.action('DocumentFolder', 'getDocumentFolderList')

